from __future__ import annotations

import glob
import os
import pathlib

import procrunner
import pytest

from cctbx import crystal
from dxtbx.model.experiment_list import ExperimentList, ExperimentListFactory
from dxtbx.serialize import load

from dials.command_line import cluster_unit_cell


def test_dials_cluster_unit_cell_command_line(dials_regression, tmp_path):
    pytest.importorskip("scipy")
    pytest.importorskip("xfel")

    data_dir = (
        pathlib.Path(dials_regression) / "refinement_test_data" / "multi_narrow_wedges"
    )
    experiments = list(data_dir.glob("data/sweep_*/experiments.json"))

    result = procrunner.run(
        command=["dials.cluster_unit_cell", "plot.show=False"] + experiments,
        print_stdout=False,
        working_directory=tmp_path,
    )
    assert not result.returncode
    assert tmp_path.joinpath("cluster_unit_cell.png").is_file()


def test_dials_cluster_unit_cell_command_line_output_files(dials_regression, tmp_path):
    pytest.importorskip("scipy")
    pytest.importorskip("xfel")

    data_dir = (
        pathlib.Path(dials_regression) / "refinement_test_data" / "multi_narrow_wedges"
    )
    experiments = list(data_dir.glob("data/sweep_*/experiments.json"))
    reflections = list(data_dir.glob("data/sweep_*/reflections.pickle"))

    # Combine experiments. Write PHIL file to avoid "command line is too long" error on Windows
    with open(tmp_path / "input.phil", "w") as f:
        f.writelines((f"input.reflections={i}" + os.linesep for i in reflections))
        f.writelines((f"input.experiments={i}" + os.linesep for i in experiments))
    result = procrunner.run(
        command=["dials.combine_experiments", "input.phil"],
        working_directory=tmp_path,
    )

    assert not result.returncode
    assert (tmp_path / "combined.refl").is_file()
    assert (tmp_path / "combined.expt").is_file()

    result = procrunner.run(
        command=[
            "dials.cluster_unit_cell",
            "plot.show=False",
            tmp_path / "combined.refl",
            tmp_path / "combined.expt",
            "output.clusters=True",
            "threshold=40",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode
    assert (tmp_path / "cluster_unit_cell.png").is_file()
    assert (tmp_path / "cluster_0.refl").is_file()
    assert (tmp_path / "cluster_0.expt").is_file()
    expts = load.experiment_list(tmp_path / "cluster_0.expt", check_format=False)
    assert len(expts) == 101
    assert (tmp_path / "cluster_1.refl").is_file()
    assert (tmp_path / "cluster_1.expt").is_file()
    expts = load.experiment_list(tmp_path / "cluster_1.expt", check_format=False)
    assert len(expts) == 1
    assert (tmp_path / "cluster_2.refl").is_file()
    assert (tmp_path / "cluster_2.expt").is_file()
    expts = load.experiment_list(tmp_path / "cluster_2.expt", check_format=False)
    assert len(expts) == 1

    result = procrunner.run(
        command=[
            "dials.split_experiments",
            tmp_path / "combined.refl",
            tmp_path / "combined.expt",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode
    experiments = list(tmp_path.glob("split_*.expt"))
    reflections = list(tmp_path.glob("split_*.refl"))

    # Write PHIL file to avoid "command line is too long" error on Windows
    with open(tmp_path / "input.phil", "w") as f:
        f.writelines((f"input.reflections={i}" + os.linesep for i in reflections))
        f.writelines((f"input.experiments={i}" + os.linesep for i in experiments))
    result = procrunner.run(
        command=[
            "dials.cluster_unit_cell",
            "output.clusters=True",
            "threshold=40",
            "plot.show=False",
            "input.phil",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode


def test_cluster_unit_cell_api(dials_regression):
    pytest.importorskip("scipy")
    pytest.importorskip("xfel")

    data_dir = os.path.join(
        dials_regression, "refinement_test_data", "multi_narrow_wedges"
    )
    experiments = ExperimentList(
        [
            ExperimentListFactory.from_json_file(expt, check_format=False)[0]
            for expt in glob.glob(
                os.path.join(data_dir, "data/sweep_*/experiments.json")
            )
        ]
    )
    crystal_symmetries = [
        crystal.symmetry(
            unit_cell=expt.crystal.get_unit_cell(),
            space_group=expt.crystal.get_space_group(),
        )
        for expt in experiments
    ]

    params = cluster_unit_cell.phil_scope.extract()
    params.plot.show = False
    params.plot.name = None
    clusters = cluster_unit_cell.do_cluster_analysis(crystal_symmetries, params)
    assert len(clusters) == 1
    cluster = clusters[0]
    assert len(cluster) == 40
    assert cluster.median_cell == pytest.approx(
        [
            90.9430182020995,
            90.9430182020995,
            90.9430182020995,
            109.47122063449069,
            109.47122063449069,
            109.47122063449069,
        ],
        abs=1e-6,
    )
    assert cluster.cell_std == pytest.approx(
        [0.09509739126548639, 0.09509739126548526, 0.0950973912654865, 0, 0, 0],
        abs=1e-6,
    )



"""
This program performs scaling on integrated datasets, which attempts to improve
the internal consistency of the reflection intensities by correcting for
various experimental effects. By default, a physically motivated scaling model
is used, with a scale, decay (B-factor) and absorption correction.
If the input files contain multiple datasets, all data will be scaled against
a common target of unique reflection intensities.

The program outputs one scaled.refl file, which contains updated reflection
intensities, variances and per-refelction scale factors, and one scaled.expt
containing the scaling models. These values can then be used to merge the data
with dials.merge for downstream structural solution. Alternatively, the
scaled.expt and scaled.refl files can be passed back to dials.scale, and
further scaling will be performed, starting from where the previous job finished.

A dials.scale.html file is also generated, containing interactive plots of merging
statistics and scaling model plots.

Example use cases

Regular single-sequence scaling, with no absorption correction::

  dials.scale integrated.refl integrated.expt physical.absorption_correction=False

Scaling multiple datasets, specifying a resolution limit::

  dials.scale 1_integrated.refl 1_integrated.expt 2_integrated.refl 2_integrated.expt d_min=1.4

Incremental scaling (with different options per dataset)::

  dials.scale integrated.refl integrated.expt physical.scale_interval=10.0

  dials.scale integrated_2.refl integrated_2.expt scaled.refl scaled.expt physical.scale_interval=15.0

More detailed documentation on usage of dials.scale can be found in the
`dials scale user guide <https://dials.github.io/dials_scale_user_guide.html>`_
"""

from __future__ import annotations

import logging
import sys
from io import StringIO

from libtbx import phil

from dials.algorithms.scaling.algorithm import ScaleAndFilterAlgorithm, ScalingAlgorithm
from dials.util import Sorry, log, show_mail_handle_errors
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

try:
    from typing import List
except ImportError:
    pass

logger = logging.getLogger("dials")
phil_scope = phil.parse(
    """
  include scope dials.algorithms.scaling.model.model.model_phil_scope
  output {
    log = dials.scale.log
      .type = str
      .help = "The log filename"
    experiments = "scaled.expt"
      .type = str
      .help = "Option to set filepath for output experiments."
    reflections = "scaled.refl"
      .type = str
      .help = "Option to set filepath for output intensities."
    html = "dials.scale.html"
      .type = str
      .help = "Filename for html report."
    json = None
      .type = str
      .help = "Filename to save html report data in json format."
    unmerged_mtz = None
      .type = str
      .help = "Filename to export an unmerged_mtz file using dials.export."
    merged_mtz = None
      .type = str
      .help = "Filename to export a merged_mtz file."
    crystal_name = XTAL
      .type = str
      .help = "The crystal name to be exported in the mtz file metadata"
      .expert_level = 1
    project_name = DIALS
      .type = str
      .help = "The project name for the mtz file metadata"
    use_internal_variance = False
      .type = bool
      .help = "Option to use internal spread of the intensities when merging
              reflection groups and calculating sigI, rather than using the
              sigmas of the individual reflections."
      .expert_level = 1
    merging.nbins = 20
      .type = int
      .help = "Number of bins to use for calculating and plotting merging stats."
      .expert_level = 1
    delete_integration_shoeboxes = True
      .type = bool
      .help = "Discard integration shoebox data from scaling output, to help"
              "with memory management."
      .expert_level = 2
  }
  include scope dials.algorithms.scaling.scaling_options.phil_scope
  include scope dials.algorithms.scaling.cross_validation.cross_validate.phil_scope
  include scope dials.algorithms.scaling.scaling_refiner.scaling_refinery_phil_scope
  include scope dials.algorithms.scaling.scale_and_filter.phil_scope
  include scope dials.util.exclude_images.phil_scope
  include scope dials.util.multi_dataset_handling.phil_scope
""",
    process_includes=True,
)


def _export_merged_mtz(params, experiments, joint_table):
    """Export merged data."""
    # call dials.merge
    from dials.command_line.merge import merge_data_to_mtz
    from dials.command_line.merge import phil_scope as merge_phil_scope

    merge_params = merge_phil_scope.extract()
    logger.disabled = True
    merge_params.assess_space_group = False
    merge_params.partiality_threshold = params.cut_data.partiality_cutoff
    merge_params.output.crystal_names = [params.output.crystal_name]
    merge_params.output.project_name = params.output.project_name
    merge_params.output.html = None
    merge_params.best_unit_cell = params.reflection_selection.best_unit_cell
    mtz_file = merge_data_to_mtz(merge_params, experiments, [joint_table])
    logger.disabled = False
    logger.info("\nWriting merged data to %s", (params.output.merged_mtz))
    out = StringIO()
    mtz_file.show_summary(out=out)
    logger.info(out.getvalue())
    mtz_file.write(params.output.merged_mtz)


def _export_unmerged_mtz(params, experiments, reflection_table):
    """Export data to unmerged_mtz format."""
    from dials.command_line.export import export_mtz
    from dials.command_line.export import phil_scope as export_phil_scope

    export_params = export_phil_scope.extract()
    export_params.intensity = ["scale"]
    export_params.mtz.partiality_threshold = params.cut_data.partiality_cutoff
    export_params.mtz.crystal_name = params.output.crystal_name
    export_params.mtz.project_name = params.output.project_name
    export_params.mtz.best_unit_cell = params.reflection_selection.best_unit_cell
    if params.cut_data.d_min:
        export_params.mtz.d_min = params.cut_data.d_min
    logger.info(
        "\nSaving output to an unmerged mtz file to %s.", params.output.unmerged_mtz
    )
    export_params.mtz.hklout = params.output.unmerged_mtz
    export_mtz(export_params, experiments, [reflection_table])


def run_scaling(params, experiments, reflections):
    """Run scaling algorithms; cross validation, scaling + filtering or standard.

    Returns:
        experiments: an experiment list with scaled data (if created)
        joint_table: a single reflection table containing scaled data (if created).
    """

    if params.output.delete_integration_shoeboxes:
        for r in reflections:
            del r["shoebox"]

    if params.cross_validation.cross_validation_mode:
        from dials.algorithms.scaling.cross_validation.cross_validate import (
            cross_validate,
        )
        from dials.algorithms.scaling.cross_validation.crossvalidator import (
            DialsScaleCrossValidator,
        )

        cross_validator = DialsScaleCrossValidator(experiments, reflections)
        cross_validate(params, cross_validator)

        logger.info(
            "Cross validation analysis does not produce scaling output files, rather\n"
            "it gives insight into the dataset. Choose an appropriate parameterisation\n"
            "and rerun scaling without cross_validation_mode.\n"
        )
        return (None, None)

    else:
        if params.filtering.method:
            algorithm = ScaleAndFilterAlgorithm(params, experiments, reflections)
        else:
            algorithm = ScalingAlgorithm(params, experiments, reflections)

        algorithm.run()

        experiments, joint_table = algorithm.finish()

        return experiments, joint_table


@show_mail_handle_errors()
def run(args: List[str] = None, phil: phil.scope = phil_scope) -> None:
    """Run the scaling from the command-line."""
    usage = """Usage: dials.scale integrated.refl integrated.expt
[integrated.refl(2) integrated.expt(2) ....] [options]"""

    parser = ArgumentParser(
        usage=usage,
        read_experiments=True,
        read_reflections=True,
        phil=phil,
        check_format=False,
        epilog=__doc__,
    )
    params, options = parser.parse_args(args=args, show_diff_phil=False)

    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        sys.exit()

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    log.config(verbosity=options.verbose, logfile=params.output.log)
    logger.info(dials_version())

    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    try:
        scaled_experiments, joint_table = run_scaling(params, experiments, reflections)
    except ValueError as e:
        raise Sorry(e)
    else:
        # Note, cross validation mode does not produce scaled datafiles
        if scaled_experiments and joint_table:
            logger.info(
                "Saving the scaled experiments to %s", params.output.experiments
            )
            scaled_experiments.as_file(params.output.experiments)
            logger.info(
                "Saving the scaled reflections to %s", params.output.reflections
            )
            joint_table.as_file(params.output.reflections)

            if params.output.unmerged_mtz:
                _export_unmerged_mtz(params, scaled_experiments, joint_table)

            if params.output.merged_mtz:
                _export_merged_mtz(params, scaled_experiments, joint_table)

    logger.info(
        "See dials.github.io/dials_scale_user_guide.html for more info on scaling options"
    )


if __name__ == "__main__":
    run()



from __future__ import annotations

from dials.array_family.flex import reflection_table
from dials.model.data import MultiPanelImageVolume
from dials_algorithms_integration_sum_ext import *  # noqa: F403; lgtm
from dials_algorithms_integration_sum_ext import sum_image_volume

__all__ = (  # noqa: F405
    "SummationDouble",
    "SummationFloat",
    "integrate_by_summation",
    "sum_image_volume",
    "sum_integrate_and_update_table",
)


def sum_integrate_and_update_table(
    reflections: reflection_table, image_volume: MultiPanelImageVolume = None
) -> reflection_table:
    """Perform 3D summation integration and update a reflection table.

    Arguments:
        reflections: The reflections to integrate

    Returns:
        The integrated reflections
    """

    # Integrate and return the reflections
    if image_volume is None:
        intensity = reflections["shoebox"].summed_intensity()
    else:
        intensity = sum_image_volume(reflections, image_volume)
    reflections["intensity.sum.value"] = intensity.observed_value()
    reflections["intensity.sum.variance"] = intensity.observed_variance()
    reflections["background.sum.value"] = intensity.background_value()
    reflections["background.sum.variance"] = intensity.background_variance()
    success = intensity.observed_success()
    reflections.set_flags(success, reflections.flags.integrated_sum)
    return success



"""
Functions for handling exclusion of images ranges from scans based on a
command line option, as well as obtaining a selection to use for selecting
the corresponding reflections.
"""

from __future__ import annotations

from orderedset import OrderedSet

import iotbx.phil

from dials.array_family import flex

phil_scope = iotbx.phil.parse(
    """
  exclude_images = None
    .type = strings
    .multiple = True
    .help = "Input in the format exp:start:end"
            "Exclude a range of images (start, stop) from the dataset with"
            "experiment identifier exp  (inclusive of frames start, stop)."
            "Multiple ranges can be given in one go, e.g."
            "exclude_images=0:150:200,1:200:250"
            "exclude_images='0:150:200 1:200:250'"
    .short_caption = "Exclude images"
    .expert_level = 1
"""
)


def exclude_image_ranges_from_scans(reflections, experiments, exclude_images):
    """Using an exclude_images phil option, modify the valid image ranges for each
    experiment in the scan (if it exists). Requires experiment identifiers to be
    set already."""
    experiments = set_initial_valid_image_ranges(experiments)
    ranges_to_remove = _parse_exclude_images_commands(
        exclude_images, experiments, reflections
    )
    experiments = _remove_ranges_from_valid_image_ranges(experiments, ranges_to_remove)
    return experiments


def get_valid_image_ranges(experiments):
    """Extract valid image ranges from experiments, returning None if no scan"""
    valid_images_ranges = []
    for exp in experiments:
        if exp.scan:
            valid_images_ranges.append(exp.scan.get_valid_image_ranges(exp.identifier))
        else:
            valid_images_ranges.append(None)
    return valid_images_ranges


def set_initial_valid_image_ranges(experiments):
    """Set the valid_image_range for each experiment to be the scan image range.
    Kept separate from dxtbx.scan object as requires experiment indentifiers.
    Also this function can be called for a mix of sequences and scanless experiments.
    """
    for exp in experiments:
        if exp.scan:
            if not exp.scan.get_valid_image_ranges(exp.identifier):
                exp.scan.set_valid_image_ranges(
                    exp.identifier, [exp.scan.get_image_range()]
                )
    return experiments


def get_selection_for_valid_image_ranges(reflection_table, experiment):
    """Determine a selection for the reflection table corresponding to reflections
    that are located in valid image ranges (according to zobs.px.value)."""
    if experiment.scan:
        valid_ranges = experiment.scan.get_valid_image_ranges(experiment.identifier)
        if valid_ranges:
            valid_mask = flex.bool(reflection_table.size(), False)
            z = reflection_table["xyzobs.px.value"].parts()[2]
            for valid_range in valid_ranges:  # i in range(0, len(valid_ranges), 2):
                mask1 = z >= valid_range[0] - 1.0  # -1.0 as image valid_ranges[i] \
                # has z is valid_ranges[i] - 1 to valid_ranges[i]
                mask2 = z <= valid_range[1]
                valid_mask.set_selected(mask1 & mask2, True)
            return valid_mask
    return flex.bool(reflection_table.size(), True)  # else say all valid


def _parse_exclude_images_commands(commands, experiments, reflections):
    """Parse a list of list of command line options.

    e.g. commands = [['1:101:200'], ['0:201:300']]
    or commands = [[101:200]] allowable for a single experiment.
    builds and returns a list of tuples (exp_id, (start, stop))


    """
    ranges_to_remove = []
    for com in commands:
        if len(com) > 1:
            sub = com
        elif "," in com[0]:
            sub = com[0].split(",")
        else:
            sub = com
        for subvals in sub:
            vals = subvals.split(":")
            if len(vals) == 2:
                if len(experiments) > 1:
                    raise ValueError(
                        "Exclude images must be in the form experimentnumber:start:stop for multiple experiments"
                    )
                else:
                    ranges_to_remove.append(
                        (experiments[0].identifier, (int(vals[0]), int(vals[1])))
                    )
            else:
                if len(vals) != 3:
                    raise ValueError(
                        "Exclude images must be input in the form experimentnumber:start:stop, or start:stop for a single experiment"
                        + "Multiple ranges can be specified by comma or space separated values e.g 0:100:150,1:120:200"
                    )
                dataset_id = int(vals[0])
                for table in reflections:
                    if dataset_id in table.experiment_identifiers():
                        expid = table.experiment_identifiers()[dataset_id]
                        ranges_to_remove.append((expid, (int(vals[1]), int(vals[2]))))
                        break
    return ranges_to_remove


def _remove_ranges_from_valid_image_ranges(experiments, ranges_to_remove):
    """Update the valid_image_ranges for each experiment in the scans. Uses set
    arithmetic to determine image ranges to keep and sets a new list of ranges in
    the scan.valid_image_ranges dict."""
    ids = list(experiments.identifiers())
    for r in ranges_to_remove:
        idx = ids.index(r[0])
        exp = experiments[idx]
        if not exp.scan:
            raise ValueError("Trying to exclude a scanless experiment")
        current_range = exp.scan.get_valid_image_ranges(
            exp.identifier
        )  # list of tuples
        # use set arithmetic on image numbers to work out images to keep
        exclude = OrderedSet(range(r[1][0], r[1][1] + 1))
        current_sets = [
            OrderedSet(range(current_range[i][0], current_range[i][1] + 1))
            for i in range(0, len(current_range))
        ]
        current_images = OrderedSet()
        for s in current_sets:
            current_images = current_images | s
        new_valid_images = list(current_images - exclude)
        # now convert images to ranges for storing in the scan
        if new_valid_images:
            new_valid_ranges = [new_valid_images[0]]
            previous = new_valid_images[0]
            for i in new_valid_images[1:]:
                if i - previous > 1:
                    new_valid_ranges.append(previous)
                    new_valid_ranges.append(i)
                previous = i
            # add end one
            new_valid_ranges.append(new_valid_images[-1])
            assert len(new_valid_ranges) % 2 == 0
            # convert to list of tuples
            valid_ranges = [
                (new_valid_ranges[i], new_valid_ranges[i + 1])
                for i in range(0, len(new_valid_ranges), 2)
            ]
        else:
            valid_ranges = []
        exp.scan.set_valid_image_ranges(exp.identifier, valid_ranges)
    return experiments


"""
Functions for scaling
"""


def exclude_image_ranges_for_scaling(reflections, experiments, exclude_images):
    """Set the initial valid ranges, then exclude in the exp.scan and set
    user_excluded_in_scaling flags."""
    experiments = exclude_image_ranges_from_scans(
        reflections, experiments, exclude_images
    )
    for refl, exp in zip(reflections, experiments):
        sel = get_selection_for_valid_image_ranges(refl, exp)
        refl.set_flags(~sel, refl.flags.user_excluded_in_scaling)
        refl.unset_flags(~sel, refl.flags.scaled)
    return reflections, experiments



"""
Methods for filtering reflection tables for bad data.

The filtering methods are combined into functions to perform the relevant
filtering on a reflection table(s), to produce a filtered reflection table
ready for export or further processing.

The set of classes defined in this module have filtering methods implemented as
classmethods/staticmethods, to allow easy use of individual methods. The
different classes are to handle filtering of different intensity types - profile,
scale, sum, profile + sum, etc. All functions and classmethods/staticmethods act on
a reflection table, returning a reflection table that is typically a new object,
due to the use of flex selections. Each filtering method raises a ValueError if
no reflections remain after filtering.

Functions:
  - filter_reflection_table:
      performs a full filtering algorithm for a given intensity choice
  - sum_partial_reflections:
      combines matching partials, replacing them with a single combined value

  filter_reflection_table takes in the following parameters: min_isigi=float,
  filter_ice_rings=bool, combine_partials=bool, partiality_threshold=float,
  intensity_choice=strings (passed in as a list e.g. ['sum', 'profile'])

Classes:
  - FilteringReductionMethods:
      a collection of staticmethods applicable to any kind of intensity values
  - FilterForExportAlgorithm:
      defines a full, general filtering algorithm for a reflection table
  - PrfIntensityReducer:
      implements methods specific to filtering of profile fitted (prf) intensities
  - SumIntensityReducer
      implements methods specific to filtering of summation (sum) intensities
  - SumAndPrfIntensityReducer
      implements filtering methods when using prf intensities if present, else
      sum (per reflection)
  - ScaleIntensityReducer
      implements filtering methods for intensities output from scaling
  - AllSumPrfScaleIntensityReducer
      implements filtering methods for using all of prf, sum and scale intensities
"""

from __future__ import annotations

import logging
from collections import defaultdict
from typing import Any, List, Type

from cctbx import crystal, miller

from dials.algorithms.scaling.outlier_rejection import reject_outliers
from dials.array_family import flex
from dials.util import tabulate
from dials.util.batch_handling import assign_batches_to_reflections

logger = logging.getLogger("dials")


class NoProfilesException(Exception):
    """Custom exception when no integrated_prf reflections found."""

    pass


def filter_reflection_table_selection(
    reflection_table, intensity_choice, *args, **kwargs
):
    """Return the selection mask for filtering."""
    reflection_table["original_index"] = flex.size_t(range(0, reflection_table.size()))
    sel = flex.bool(reflection_table.size(), False)
    filtered_table = filter_reflection_table(
        reflection_table, intensity_choice, *args, **kwargs
    )
    sel.set_selected(filtered_table["original_index"], True)
    return sel


def filter_reflection_table(
    reflection_table: flex.reflection_table,
    intensity_choice: List[str],
    *args: Any,
    **kwargs: Any,
) -> flex.reflection_table:
    """Filter the data and delete unneeded intensity columns.

    A list of which intensities to filter on e.g "sum", "scale", "profile" or
    allowed combinations. If a combination is given, only those reflections
    which have valid reflections for the multiple intensity types are retained.

    Strict checks are made that the requested intensity choice(s) has the
    required data in the reflection table.

    Args:
        reflection_table: a single reflection table object
        intensity_choice[List]: a list of the which intensities to filter on

    Returns:
        A reflection table filtered based on the arguments (of reduced size
        compared to the input table.)

    Raises:
        ValueError: if invalid intensity_choice given, if one step of filtering
            causes no reflections to remain, if no profile reflections remain
            after filtering and the choice is "profile".
    """
    if not isinstance(intensity_choice, list):
        raise ValueError("intensity_choice must be List[str]")

    if intensity_choice == ["scale"]:
        reducer: Type[FilterForExportAlgorithm] = ScaleIntensityReducer
    elif intensity_choice == ["sum"]:
        reducer = SumIntensityReducer
    elif intensity_choice == ["profile"]:
        reducer = PrfIntensityReducer
    elif all(i in intensity_choice for i in ["sum", "scale", "profile"]):
        reducer = AllSumPrfScaleIntensityReducer
    elif all(i in intensity_choice for i in ["sum", "profile"]):
        reducer = SumAndPrfIntensityReducer
    elif all(i in intensity_choice for i in ["sum | profile"]):
        reducer = SumORPrfIntensityReducer
    elif all(i in intensity_choice for i in ["sum", "scale"]):
        reducer = SumAndScaleIntensityReducer
    else:
        raise ValueError(
            (
                "Unrecognised intensity choice for filter_reflection_table,\n"
                "value read: {}\n"
                "must be one of: 'scale', 'profile', 'sum', 'profile sum', \n"
                "                'sum scale', 'profile sum scale'\n"
                "(if parsing from command line, multiple choices passed as e.g. profile+sum"
            ).format(intensity_choice)
        )
    # Validate that the reflection table has the columns we need
    required_columns_lookup = {
        "scale": {"inverse_scale_factor", "intensity.scale.value"},
        "profile": {"intensity.prf.value"},
        "sum": {"intensity.sum.value"},
    }
    for intensity_kind, required_columns in required_columns_lookup.items():
        if intensity_kind in intensity_choice:
            missing_columns = required_columns - set(reflection_table.keys())
            if missing_columns:
                raise ValueError(
                    "Cannot export intensity kind '{}'; missing column(s): {}".format(
                        intensity_kind, ", ".join(missing_columns)
                    )
                )

    # Do the filtering, but with an exception for the case of no profile fitted
    # reflections - in this case, try to reprocess without profile fitted.
    try:
        reflection_table = reducer.filter_for_export(reflection_table, *args, **kwargs)
    except NoProfilesException as e:
        logger.warning(e, exc_info=True)
        if "profile" in intensity_choice:
            intensity_choice.remove("profile")
        else:
            intensity_choice = None
        if intensity_choice:
            logger.info(
                "Attempting to reprocess with intensity choice: %s",
                " + ".join(i for i in intensity_choice),
            )
            reflection_table = filter_reflection_table(
                reflection_table, intensity_choice, *args, **kwargs
            )
        else:
            raise ValueError(
                "Unable to process data due to absence of profile fitted reflections"
            )
    return reflection_table


def filtered_arrays_from_experiments_reflections(
    experiments,
    reflections,
    outlier_rejection_after_filter=False,
    partiality_threshold=0.99,
    return_batches=False,
):
    """Create a list of filtered arrays from experiments and reflections.

    A partiality threshold can be set, and if outlier_rejection_after_filter
    is True, and intensity.scale values are not present, then a round of
    outlier rejection will take place.

    Raises:
        ValueError: if no datasets remain after filtering.
    """
    miller_arrays = []
    ids_to_del = []

    if return_batches:
        assert all(expt.scan is not None for expt in experiments)
        batch_offsets = [expt.scan.get_batch_offset() for expt in experiments]
        reflections = assign_batches_to_reflections(reflections, batch_offsets)
        batch_arrays = []

    for idx, (expt, refl) in enumerate(zip(experiments, reflections)):
        crystal_symmetry = crystal.symmetry(
            unit_cell=expt.crystal.get_unit_cell(),
            space_group=expt.crystal.get_space_group(),
        )

        # want to use scale intensities if present, else sum + prf (if available)
        if "intensity.scale.value" in refl:
            intensity_choice = ["scale"]
            intensity_to_use = "intensity.scale"
        else:
            assert "intensity.sum.value" in refl
            intensity_to_use = "intensity.sum"
            intensity_choice = ["sum"]
            if "intensity.prf.value" in refl:
                intensity_choice.append("profile")
                intensity_to_use = "intensity.prf"

        try:
            logger.info("Filtering reflections for dataset %s", idx)
            refl = filter_reflection_table(
                refl,
                intensity_choice,
                min_isigi=-5,
                filter_ice_rings=False,
                combine_partials=True,
                partiality_threshold=partiality_threshold,
            )
        except ValueError:
            logger.info(
                "Dataset %s removed as no reflections left after filtering", idx
            )
            ids_to_del.append(idx)
        else:
            # If scale was chosen - will return scale or have raised ValueError
            # If prf or sum, possible was no prf but want to continue.
            try:
                refl["intensity"] = refl[intensity_to_use + ".value"]
                refl["variance"] = refl[intensity_to_use + ".variance"]
            except KeyError:  # catch case where prf were removed.
                refl["intensity"] = refl["intensity.sum.value"]
                refl["variance"] = refl["intensity.sum.variance"]
            if outlier_rejection_after_filter and intensity_to_use != "intensity.scale":
                refl = reject_outliers(refl, expt, method="simple", zmax=12.0)
                refl = refl.select(~refl.get_flags(refl.flags.outlier_in_scaling))

            miller_set = miller.set(
                crystal_symmetry, refl["miller_index"], anomalous_flag=False
            )
            intensities = miller_set.array(
                data=refl["intensity"], sigmas=flex.sqrt(refl["variance"])
            )
            intensities.set_observation_type_xray_intensity()
            intensities.set_info(
                miller.array_info(source="DIALS", source_type="pickle")
            )
            miller_arrays.append(intensities)
            if return_batches:
                batch_arrays.append(
                    miller_set.array(data=refl["batch"]).set_info(intensities.info())
                )

    if not miller_arrays:
        raise ValueError(
            """No datasets remain after pre-filtering. Please check input data.
The datasets may not contain any full reflections; the command line
option partiality_threshold can be lowered to include partials."""
        )

    for id_ in ids_to_del[::-1]:
        del experiments[id_]
        del reflections[id_]

    if return_batches:
        return miller_arrays, batch_arrays
    return miller_arrays


def checkdataremains(func):
    """Decorate a filtering method, to raise a ValueError if all data filtered."""

    def wrapper(*args, **kwargs):

        reflections = func(*args, **kwargs)

        if not reflections:
            raise ValueError("All data has been filtered from the reflection table")
        return reflections

    return wrapper


class FilteringReductionMethods:
    """A collection of methods for filtering.

    Some internal methods require an 'intensity' string, which indicates which
    column to filter on. These methods can be called multiple times to filter
    on multiple intensity choices. All methods may reduce the size of the
    reflection table by deleting data.
    """

    @staticmethod
    @checkdataremains
    def _filter_on_min_isigi(reflection_table, intensity, min_isigi=None):
        if min_isigi:
            selection = (
                reflection_table["intensity." + intensity + ".value"]
                / flex.sqrt(reflection_table["intensity." + intensity + ".variance"])
            ) < min_isigi
            reflection_table.del_selected(selection)
            logger.info(
                "Removed %d %s reflections with I/Sig(I) < %s",
                selection.count(True),
                "intensity." + intensity + ".value",
                min_isigi,
            )
        return reflection_table

    @staticmethod
    @checkdataremains
    def _filter_bad_variances(reflection_table, intensity):
        """Filter reflections with a variance <= 0."""
        selection = reflection_table["intensity." + intensity + ".variance"] <= 0
        if intensity == "prf":
            selection &= reflection_table.get_flags(
                reflection_table.flags.integrated_prf
            )
        elif intensity == "sum":
            selection &= reflection_table.get_flags(
                reflection_table.flags.integrated_sum
            )
        if selection.count(True) > 0:
            reflection_table.del_selected(selection)
            logger.info(
                "Removed %d %s reflections with negative variance",
                selection.count(True),
                "intensity." + intensity + ".value",
            )
        return reflection_table

    @staticmethod
    @checkdataremains
    def calculate_lp_qe_correction_and_filter(reflection_table):
        """Calculate the lp, qe combined correction and filter bad values."""
        # FIXME errors in e.g. LP correction need to be propagated here?
        nref = reflection_table.size()
        qe = None
        if "qe" in reflection_table:
            reflection_table = reflection_table.select(reflection_table["qe"] > 0.0)
            qe = reflection_table["qe"]
        elif "dqe" in reflection_table:
            reflection_table = reflection_table.select(reflection_table["dqe"] > 0.0)
            qe = reflection_table["dqe"]
        if reflection_table.size() < nref:
            logger.info(
                "%s reflections filtered due to bad dqe/qe value",
                nref - reflection_table.size(),
            )
        # Now calculate conversion factor
        conversion = flex.double(reflection_table.size(), 1.0)
        if qe:
            conversion /= qe
        if "lp" in reflection_table:
            conversion *= reflection_table["lp"]
        return reflection_table, conversion

    @staticmethod
    @checkdataremains
    def filter_ice_rings(reflection_table):
        """Filter reflections with the in_powder_ring flag."""
        selection = reflection_table.get_flags(reflection_table.flags.in_powder_ring)
        reflection_table.del_selected(selection)
        logger.info(
            "Removed %d reflections in ice ring resolutions", selection.count(True)
        )
        return reflection_table

    @staticmethod
    @checkdataremains
    def filter_on_d_min(reflection_table, d_min):
        """Filter reflections below a d-value."""
        selection = reflection_table["d"] <= d_min
        logger.info(
            "Removed %d reflections with d <= %.2f", selection.count(True), d_min
        )
        reflection_table.del_selected(selection)
        return reflection_table

    @staticmethod
    @checkdataremains
    def filter_on_d_max(reflection_table, d_max):
        """Filter reflections above a d-value."""
        selection = reflection_table["d"] >= d_max
        logger.info(
            "Removed %d reflections with d >= %.2f", selection.count(True), d_max
        )
        reflection_table.del_selected(selection)
        return reflection_table

    @staticmethod
    @checkdataremains
    def filter_unassigned_reflections(reflection_table):
        """Remove reflections that are not assigned to an experiment."""
        reflection_table = reflection_table.select(reflection_table["id"] >= 0)
        logger.info("Read %s predicted reflections", reflection_table.size())
        return reflection_table

    @staticmethod
    @checkdataremains
    def combine_and_filter_partials(
        reflection_table, partiality_threshold, combine_partials=True
    ):
        """Combine partials and filter those below the threshold."""
        if "partiality" in reflection_table:
            reflection_table["fractioncalc"] = reflection_table["partiality"]
            if combine_partials and "partial_id" in reflection_table:
                dataset_ids = set(reflection_table["id"])
                n_datasets = len(dataset_ids)
                if n_datasets > 1:
                    total_reflection_table = flex.reflection_table()
                    for id_ in dataset_ids:
                        single_table = reflection_table.select(
                            reflection_table["id"] == id_
                        )
                        total_reflection_table.extend(
                            sum_partial_reflections(single_table)
                        )
                    reflection_table = total_reflection_table
                else:
                    reflection_table = sum_partial_reflections(reflection_table)
                reflection_table["fractioncalc"] = reflection_table["partiality"]
            selection = reflection_table["partiality"] < partiality_threshold
            if selection.count(True) > 0:
                reflection_table.del_selected(selection)
                logger.info(
                    "Removed %d reflections below partiality threshold",
                    selection.count(True),
                )
        else:
            reflection_table["fractioncalc"] = flex.double(reflection_table.size(), 1.0)
        return reflection_table


class FilterForExportAlgorithm(FilteringReductionMethods):
    """Definition of the filter_for_export algorithm.

    An abstract class, from which reduction methods for particular intensity
    types can be implemented in a subclass.
    """

    allowed_intensities = ["prf", "scale", "sum"]  # Supported intensities
    # subclasses must define a class attribute intensities, which is a list
    # of a subset of the allowed intensities.

    @classmethod
    def filter_for_export(
        cls,
        reflection_table,
        min_isigi=None,
        filter_ice_rings=False,
        combine_partials=True,
        partiality_threshold=0.99,
        d_min=None,
        d_max=None,
    ):
        """Apply the filtering methods to reflection table."""
        assert (
            reflection_table.size() > 0
        ), """Empty reflection table given to reduce_data_for_export function"""
        reflection_table = cls.filter_unassigned_reflections(reflection_table)
        reflection_table = cls.reduce_on_intensities(reflection_table)

        for intensity in cls.allowed_intensities:
            if intensity not in cls.intensities:
                if "intensity." + intensity + ".value" in reflection_table:
                    del reflection_table["intensity." + intensity + ".value"]
                    del reflection_table["intensity." + intensity + ".variance"]
            else:
                msg = "No intensity." + intensity + " values found in reflection table"
                assert "intensity." + intensity + ".value" in reflection_table, msg
                assert "intensity." + intensity + ".variance" in reflection_table, msg

        reflection_table = cls.filter_bad_variances(reflection_table)
        if filter_ice_rings:
            reflection_table = cls.filter_ice_rings(reflection_table)
        if d_min:
            reflection_table = cls.filter_on_d_min(reflection_table, d_min)
        if d_max:
            reflection_table = cls.filter_on_d_max(reflection_table, d_max)

        reflection_table = cls.apply_scaling_factors(reflection_table)

        reflection_table = cls.combine_and_filter_partials(
            reflection_table, partiality_threshold, combine_partials
        )

        # Select on I/sigI after applying scale factors and combining partials
        if min_isigi:
            reflection_table = cls.filter_on_min_isigi(reflection_table, min_isigi)

        return reflection_table

    @staticmethod
    def reduce_on_intensities(reflection_table):
        """Select reflections successfully processed by the relevant method."""
        raise NotImplementedError()

    @classmethod
    def filter_bad_variances(cls, reflection_table):
        """Remove reflection if either has a bad variance."""
        for intensity in cls.intensities:
            reflection_table = cls._filter_bad_variances(reflection_table, intensity)
        return reflection_table

    @classmethod
    def filter_on_min_isigi(cls, reflection_table, min_isigi=None):
        """Remove reflection if either has a IsigI below min_isigi."""
        for intensity in cls.intensities:
            reflection_table = cls._filter_on_min_isigi(
                reflection_table, intensity, min_isigi
            )
        return reflection_table

    @classmethod
    def apply_scaling_factors(cls, reflection_table):
        """Apply the relevant scaling factors including lp, qde, scale etc."""
        raise NotImplementedError()


class PrfIntensityReducer(FilterForExportAlgorithm):
    """Reduction methods for data with profile intensities."""

    intensities = ["prf"]

    @staticmethod
    @checkdataremains
    def reduce_on_intensities(reflection_table):
        """Select reflections successfully integrated by profile fitting.

        Raises:
            NoProfilesException: Custom Exception to indicate no reflections
                with the integrated_prf flag.
        """
        selection = reflection_table.get_flags(reflection_table.flags.integrated_prf)
        if selection.count(True) == 0:
            raise NoProfilesException(
                "WARNING: No profile-integrated reflections found"
            )
        reflection_table = reflection_table.select(selection)
        logger.info(
            "Selected %d profile integrated reflections", reflection_table.size()
        )
        return reflection_table

    @classmethod
    def apply_scaling_factors(cls, reflection_table):
        """Apply corrections to the intensities and variances (partiality, lp, qe)."""
        if "partiality" in reflection_table:
            reflection_table = reflection_table.select(
                reflection_table["partiality"] > 0.0
            )

        reflection_table, conversion = cls.calculate_lp_qe_correction_and_filter(
            reflection_table
        )

        reflection_table["intensity.prf.value"] *= conversion
        reflection_table["intensity.prf.variance"] *= conversion * conversion
        return reflection_table


class SumIntensityReducer(FilterForExportAlgorithm):
    """Reduction methods for data with sum intensities."""

    intensities = ["sum"]

    @staticmethod
    @checkdataremains
    def reduce_on_intensities(reflection_table):
        """Select reflections successfully integrated by summation method."""
        selection = reflection_table.get_flags(reflection_table.flags.integrated_sum)
        reflection_table = reflection_table.select(selection)
        logger.info(
            "Selected %d summation integrated reflections", reflection_table.size()
        )
        return reflection_table

    @classmethod
    def apply_scaling_factors(cls, reflection_table):
        """Apply corrections to the intensities and variances (partiality, lp, qe)."""
        reflection_table, conversion = cls.calculate_lp_qe_correction_and_filter(
            reflection_table
        )

        if "partiality" in reflection_table:
            nonzero_sel = reflection_table["partiality"] > 0.0
            reflection_table = reflection_table.select(nonzero_sel)
            conversion = conversion.select(nonzero_sel)
            conversion /= reflection_table["partiality"]

        reflection_table["intensity.sum.value"] *= conversion
        reflection_table["intensity.sum.variance"] *= conversion * conversion
        return reflection_table


class SumAndPrfIntensityReducer(FilterForExportAlgorithm):
    """Reduction methods for data with sum and profile intensities.

    Only reflections with valid values for all intensity types are retained.
    """

    intensities = ["sum", "prf"]

    @staticmethod
    @checkdataremains
    def reduce_on_intensities(reflection_table):
        """Select reflections successfully integrated by sum and prf methods."""
        if (
            reflection_table.get_flags(reflection_table.flags.integrated_prf).count(
                True
            )
            == 0
        ):
            raise NoProfilesException(
                "WARNING: No profile-integrated reflections found"
            )
        selection = reflection_table.get_flags(
            reflection_table.flags.integrated, all=True
        )
        reflection_table = reflection_table.select(selection)
        logger.info(
            "Selected %d reflections integrated by profile and summation methods",
            reflection_table.size(),
        )
        return reflection_table

    @classmethod
    def apply_scaling_factors(cls, reflection_table):
        """Apply corrections to the intensities and variances (partiality, lp, qe)."""
        reflection_table, conversion = cls.calculate_lp_qe_correction_and_filter(
            reflection_table
        )
        sum_conversion = conversion

        if "partiality" in reflection_table:
            nonzero_sel = reflection_table["partiality"] > 0.0
            reflection_table = reflection_table.select(nonzero_sel)
            conversion = conversion.select(nonzero_sel)
            sum_conversion = conversion / reflection_table["partiality"]

        reflection_table["intensity.sum.value"] *= sum_conversion
        reflection_table["intensity.sum.variance"] *= sum_conversion * sum_conversion
        reflection_table["intensity.prf.value"] *= conversion
        reflection_table["intensity.prf.variance"] *= conversion * conversion
        return reflection_table


class SumORPrfIntensityReducer(FilterForExportAlgorithm):
    """Reduction methods for data with sum or profile intensities.

    Reflections with valid values for either intensity type are retained.
    """

    intensities = ["sum", "prf"]

    @staticmethod
    @checkdataremains
    def reduce_on_intensities(reflection_table):
        """Select reflections successfully integrated by sum and prf methods."""
        if (
            reflection_table.get_flags(reflection_table.flags.integrated_prf).count(
                True
            )
            == 0
        ):
            raise NoProfilesException(
                "WARNING: No profile-integrated reflections found"
            )
        selection = reflection_table.get_flags(
            reflection_table.flags.integrated, all=False
        )
        reflection_table = reflection_table.select(selection)
        logger.info(
            "Selected %d reflections integrated by profile or summation methods",
            reflection_table.size(),
        )
        return reflection_table

    @classmethod
    def apply_scaling_factors(cls, reflection_table):
        """Apply corrections to the intensities and variances (partiality, lp, qe)."""
        reflection_table, conversion = cls.calculate_lp_qe_correction_and_filter(
            reflection_table
        )
        sum_conversion = conversion

        if "partiality" in reflection_table:
            nonzero_sel = reflection_table["partiality"] > 0.0
            reflection_table = reflection_table.select(nonzero_sel)
            conversion = conversion.select(nonzero_sel)
            sum_conversion = conversion / reflection_table["partiality"]

        reflection_table["intensity.sum.value"] *= sum_conversion
        reflection_table["intensity.sum.variance"] *= sum_conversion * sum_conversion
        reflection_table["intensity.prf.value"] *= conversion
        reflection_table["intensity.prf.variance"] *= conversion * conversion
        return reflection_table


class ScaleIntensityReducer(FilterForExportAlgorithm):
    """Reduction methods for data with sum intensities."""

    intensities = ["scale"]

    @staticmethod
    @checkdataremains
    def reduce_on_intensities(reflection_table):
        """Select intensities used for scaling and remove scaling outliers."""
        selection = ~reflection_table.get_flags(
            reflection_table.flags.bad_for_scaling, all=False
        )
        reflection_table = reflection_table.select(selection)
        logger.info("Selected %d scaled reflections", reflection_table.size())
        assert "inverse_scale_factor" in reflection_table
        selection = reflection_table["inverse_scale_factor"] <= 0.0
        if selection.count(True) > 0:
            reflection_table.del_selected(selection)
            logger.info(
                "Removed %s reflections with zero or negative inverse scale factors",
                selection.count(True),
            )
        return reflection_table

    @staticmethod
    def apply_scaling_factors(reflection_table):
        """Apply the inverse scale factor to the scale intensities."""
        if "partiality" in reflection_table:
            reflection_table = reflection_table.select(
                reflection_table["partiality"] > 0.0
            )

        assert "inverse_scale_factor" in reflection_table
        reflection_table["intensity.scale.value"] /= reflection_table[
            "inverse_scale_factor"
        ]
        reflection_table["intensity.scale.variance"] /= flex.pow2(
            reflection_table["inverse_scale_factor"]
        )
        return reflection_table


class AllSumPrfScaleIntensityReducer(FilterForExportAlgorithm):
    """Reduction methods for data with sum, profile and scale intensities.

    Only reflections with valid values for all intensity types are retained.
    """

    intensities = ["sum", "prf", "scale"]

    @staticmethod
    def reduce_on_intensities(reflection_table):
        """Select those with valid reflections for all values."""
        reflection_table = SumAndPrfIntensityReducer.reduce_on_intensities(
            reflection_table
        )
        reflection_table = ScaleIntensityReducer.reduce_on_intensities(reflection_table)
        return reflection_table

    @staticmethod
    def apply_scaling_factors(reflection_table):
        """Apply corrections to the intensities and variances."""
        reflection_table = SumAndPrfIntensityReducer.apply_scaling_factors(
            reflection_table
        )
        reflection_table = ScaleIntensityReducer.apply_scaling_factors(reflection_table)
        return reflection_table


class SumAndScaleIntensityReducer(FilterForExportAlgorithm):
    """Reduction methods for data with sum and scale intensities.

    Only reflections with valid values for all intensity types are retained.
    """

    intensities = ["sum", "scale"]

    @staticmethod
    def reduce_on_intensities(reflection_table):
        """Select those with valid reflections for all values."""
        reflection_table = SumIntensityReducer.reduce_on_intensities(reflection_table)
        reflection_table = ScaleIntensityReducer.reduce_on_intensities(reflection_table)
        return reflection_table

    @staticmethod
    def apply_scaling_factors(reflection_table):
        """Apply corrections to the intensities and variances."""
        reflection_table = SumIntensityReducer.apply_scaling_factors(reflection_table)
        reflection_table = ScaleIntensityReducer.apply_scaling_factors(reflection_table)
        return reflection_table


def sum_partial_reflections(reflection_table):
    """Sum partial reflections if more than one recording of a reflection present.

    This is a weighted sum for summation integration; weighted average for
    profile fitted reflections. N.B. this will report total partiality for
    the summed reflection.
    """

    if ("partiality" not in reflection_table) or ("partial_id" not in reflection_table):
        return reflection_table

    nrefl = reflection_table.size()
    intensities = []
    for intensity in ["prf", "scale", "sum"]:
        if "intensity." + intensity + ".value" in reflection_table:
            intensities.append(intensity)

    sel = reflection_table["partiality"] < 0.99
    if not sel.count(True):
        return reflection_table

    for intensity in intensities:
        sel = sel & (reflection_table["intensity." + intensity + ".variance"] > 0)
    isel = sel.iselection()

    # create map of partial_id to reflections
    delete = flex.size_t()
    partial_map = defaultdict(list)
    for j in isel:
        partial_map[reflection_table["partial_id"][j]].append(j)

    # now work through this map - get total partiality for every reflection;
    # here only consider reflections with > 1 component;
    partial_ids = []
    for p_id in partial_map:
        if len(partial_map[p_id]) > 1:
            partial_ids.append(p_id)

    header = ["Partial id", "Partiality"]
    for i in intensities:
        header.extend([str(i) + " intensity", str(i) + " variance"])
    rows = []

    # Now loop through 'matched' partials, summing and then deleting before return
    for p_id in partial_ids:
        j = partial_map[p_id]
        for i in j:
            data = [str(p_id), str(reflection_table["partiality"][i])]
            for intensity in intensities:
                data.extend(
                    [
                        str(reflection_table["intensity." + intensity + ".value"][i]),
                        str(
                            reflection_table["intensity." + intensity + ".variance"][i]
                        ),
                    ]
                )
            rows.append(data)

        # do the summing of the partiality values separately to allow looping
        # over multiple times
        total_partiality = sum(reflection_table["partiality"][i] for i in j)
        if "prf" in intensities:
            reflection_table = _sum_prf_partials(reflection_table, j)
        if "sum" in intensities:
            reflection_table = _sum_sum_partials(reflection_table, j)
        if "scale" in intensities:
            reflection_table = _sum_scale_partials(reflection_table, j)
        # FIXME now that the partials have been summed, should fractioncalc be set
        # to one (except for summation case?)
        reflection_table["partiality"][j[0]] = total_partiality
        delete.extend(flex.size_t(j[1:]))
        data = ["combined " + str(p_id), str(total_partiality)]
        for intensity in intensities:
            data.extend(
                [
                    str(reflection_table["intensity." + intensity + ".value"][j[0]]),
                    str(reflection_table["intensity." + intensity + ".variance"][j[0]]),
                ]
            )
        rows.append(data)
    reflection_table.del_selected(delete)
    if nrefl > reflection_table.size():
        logger.info(
            "Combined %s partial reflections with other partial reflections",
            nrefl - reflection_table.size(),
        )

    # Formatting this table can be sloooow for large numbers of reflections, so skip
    # this unless debug output has been requested
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug("\nSummary of combination of partial reflections")
        logger.debug(tabulate(rows, header))
    return reflection_table


# FIXME what are the correct weights to use for the different cases? - why
# weighting by (I/sig(I))^2 not just 1/variance for prf. See tests?


def _sum_prf_partials(reflection_table, partials_isel_for_pid):
    """Sum prf partials and set the updated value in the first entry."""
    j = partials_isel_for_pid
    value = reflection_table["intensity.prf.value"][j[0]]
    variance = reflection_table["intensity.prf.variance"][j[0]]
    weight = value * value / variance
    value *= weight
    variance *= weight
    total_weight = weight
    for i in j[1:]:
        _value = reflection_table["intensity.prf.value"][i]
        _variance = reflection_table["intensity.prf.variance"][i]
        _weight = _value * _value / _variance
        value += _weight * _value
        variance += _weight * _variance
        total_weight += _weight
    # now write these back into original reflection
    if total_weight:
        reflection_table["intensity.prf.value"][j[0]] = value / total_weight
        reflection_table["intensity.prf.variance"][j[0]] = variance / total_weight
    else:
        variance = sum(reflection_table["intensity.prf.variance"][_j] for _j in j)
        reflection_table["intensity.prf.value"][j[0]] = 0
        reflection_table["intensity.prf.variance"][j[0]] = variance
    return reflection_table


def _sum_sum_partials(reflection_table, partials_isel_for_pid):
    """Sum sum partials and set the updated value in the first entry."""
    j = partials_isel_for_pid
    value = reflection_table["intensity.sum.value"][j[0]]
    variance = reflection_table["intensity.sum.variance"][j[0]]
    for i in j[1:]:
        value += reflection_table["intensity.sum.value"][i]
        variance += reflection_table["intensity.sum.variance"][i]
    reflection_table["intensity.sum.value"][j[0]] = value
    reflection_table["intensity.sum.variance"][j[0]] = variance
    return reflection_table


def _sum_scale_partials(reflection_table, partials_isel_for_pid):
    """Sum scale partials and set the updated value in the first entry."""
    # Weight scaled intensity partials by 1/variance. See
    # https://en.wikipedia.org/wiki/Weighted_arithmetic_mean, section
    # 'Dealing with variance'
    j = partials_isel_for_pid
    variance = reflection_table["intensity.scale.variance"][j[0]]
    value = reflection_table["intensity.scale.value"][j[0]] / variance
    total_weight = 1.0 / variance
    for i in j[1:]:
        _variance = reflection_table["intensity.scale.variance"][i]
        value += reflection_table["intensity.scale.value"][i] / _variance
        total_weight += 1.0 / _variance
    reflection_table["intensity.scale.value"][j[0]] = value / total_weight
    reflection_table["intensity.scale.variance"][j[0]] = 1.0 / total_weight
    return reflection_table



"""
Algorithms for analysis of resolution limits.
"""

from __future__ import annotations

import enum
import logging
import math
import typing

import iotbx.merging_statistics
import iotbx.mtz
import iotbx.phil
from cctbx import miller, uctbx
from cctbx.array_family import flex
from iotbx.reflection_file_utils import label_table
from scitbx.math import curve_fitting, five_number_summary

from dials.algorithms.scaling.scaling_library import determine_best_unit_cell
from dials.report import plots
from dials.util import Sorry, tabulate
from dials.util.batch_handling import (
    assign_batches_to_reflections,
    calculate_batch_offsets,
)
from dials.util.filter_reflections import filter_reflection_table
from dials.util.normalisation import quasi_normalisation

logger = logging.getLogger(__name__)


class metrics(enum.Enum):
    """Supported metrics for estimation of resolution limits."""

    CC_HALF = "cc_half"
    CC_REF = "cc_ref"
    ISIGMA = "unmerged_i_over_sigma_mean"
    MISIGMA = "i_over_sigma_mean"
    I_MEAN_OVER_SIGMA_MEAN = "i_mean_over_sigi_mean"
    RMERGE = "r_merge"
    COMPLETENESS = "completeness"


def polynomial_fit(x, y, degree=5):
    """
    Fit a polynomial to the values y(x) and return this fit

    x, y should be iterables containing floats of the same size. The order is the order
    of polynomial to use for this fit. This will be useful for e.g. I/sigma.
    """

    fit = curve_fitting.univariate_polynomial_fit(
        x, y, degree=degree, max_iterations=100
    )
    f = curve_fitting.univariate_polynomial(*fit.params)
    return f(x)


def tanh_fit(x, y, iqr_multiplier=None):
    """
    Fit a tanh function to the values y(x) and return this fit

    x, y should be iterables containing floats of the same size. This is used for
    fitting a curve to CC½.
    """

    tf = curve_fitting.tanh_fit(x, y)
    f = curve_fitting.tanh(*tf.params)

    if iqr_multiplier:
        assert iqr_multiplier > 0
        yc = f(x)
        dy = y - yc

        min_x, q1_x, med_x, q3_x, max_x = five_number_summary(dy)
        iqr_x = q3_x - q1_x
        cut_x = iqr_multiplier * iqr_x
        outliers = (dy > q3_x + cut_x) | (dy < q1_x - cut_x)
        if outliers.count(True) > 0:
            xo = x.select(~outliers)
            yo = y.select(~outliers)
            tf = curve_fitting.tanh_fit(xo, yo)
            f = curve_fitting.tanh(*tf.params)

    return f(x)


def log_fit(x, y, degree=5):
    """Fit the values log(y(x)) then return exp() to this fit.

    x, y should be iterables containing floats of the same size. The order is the order
    of polynomial to use for this fit. This will be useful for e.g. I/sigma."""

    fit = curve_fitting.univariate_polynomial_fit(
        x, flex.log(y), degree=degree, max_iterations=100
    )
    f = curve_fitting.univariate_polynomial(*fit.params)
    return flex.exp(f(x))


def log_inv_fit(x, y, degree=5):
    """Fit the values log(1 / y(x)) then return the inverse of this fit.

    x, y should be iterables, the order of the polynomial for the transformed
    fit needs to be specified. This will be useful for e.g. Rmerge."""

    fit = curve_fitting.univariate_polynomial_fit(
        x, flex.log(1 / y), degree=degree, max_iterations=100
    )
    f = curve_fitting.univariate_polynomial(*fit.params)
    return 1 / flex.exp(f(x))


def resolution_fit_from_merging_stats(merging_stats, metric, model, limit, sel=None):
    """Estimate a resolution limit based on the input `metric`

    The function defined by `model` will be fit to the selected `metric` which has been
    pre-calculated by the `merging_stats` object. The estimated resolution limit is
    chosen as the `d_star_sq` value at which the fitted function equals `limit`.

    Args:
        merging_stats (iotbx.merging_statistics.dataset_statistics): Pre-calculated
            merging statistics object
        metric (str): The metric to use for estimating a resolution limit. Must be a
            metric calculated by `iotbx.merging_statistics.merging_stats` and
            available as an attribute on the `bins` attribute of the input
            `merging_stats` object.
        model: The function to fit to the selected metric. Must be callable, taking as
            input x (d_star_sq) and y (the metric to be fitted) values, returning the
            fitted y(x) values.
        limit (float): The resolution limit criterion.
        sel (scitbx.array_family.flex.bool): An optional selection to apply to the
            `merging_stats` bins.

    Returns: The estimated resolution limit in units of Å^-1
    """

    y_obs = flex.double(getattr(b, metric) for b in merging_stats.bins).reversed()
    d_star_sq = flex.double(
        uctbx.d_as_d_star_sq(b.d_min) for b in merging_stats.bins
    ).reversed()
    return resolution_fit(d_star_sq, y_obs, model, limit, sel=sel)


def resolution_fit(d_star_sq, y_obs, model, limit, sel=None):
    """Estimate a resolution limit based on the input merging statistics

    The function defined by `model` will be fit to the input `d_star_sq` and `y_obs`.
    The estimated resolution limit is chosen as the `d_star_sq` value at which the
    fitted function equals `limit`.

    Args:
        d_star_sq (scitbx.array_family.flex.double): The high resolution limits of the
            resolution bins in units 1/d*2
        y_obs (scitbx.array_family.flex.double): The statistic against which to fit the
            function `model`
        model: The function to fit against `y_obs`. Must be callable, taking as input x
            (d_star_sq) and y (the metric to be fitted) values, returning the fitted
            y(x) values.
        limit (float): The resolution limit criterion.
        sel (scitbx.array_family.flex.bool): An optional selection to apply to the
            `d_star_sq` and `y_obs` values.

    Returns: The estimated resolution limit in units of Å^-1

    Raises:
        RuntimeError: Raised if no `y_obs` values remain after application of the
        selection `sel`
    """
    if not sel:
        sel = flex.bool(len(d_star_sq), True)
    sel &= y_obs > 0
    y_obs = y_obs.select(sel)
    d_star_sq = d_star_sq.select(sel)

    if not len(y_obs):
        raise RuntimeError("No reflections left for fitting")
    y_fit = model(d_star_sq, y_obs, 6)
    logger.debug(
        tabulate(
            [("d*2", "d", "obs", "fit")]
            + [
                (ds2, uctbx.d_star_sq_as_d(ds2), yo, yf)
                for ds2, yo, yf in zip(d_star_sq, y_obs, y_fit)
            ],
            headers="firstrow",
        )
    )

    if flex.min(y_obs) > limit:
        d_min = 1.0 / math.sqrt(flex.max(d_star_sq))
    else:
        try:
            d_min = 1.0 / math.sqrt(interpolate_value(d_star_sq, y_fit, limit))
        except RuntimeError as e:
            logger.debug(f"Error interpolating value: {e}")
            d_min = None

    return ResolutionResult(d_star_sq, y_obs, y_fit, d_min)


def _get_cc_half_significance(merging_stats, cc_half_method):
    """Get the CC½ significance values from the input merging_stats object"""
    if (
        cc_half_method == "sigma_tau"
        and merging_stats.overall.cc_one_half_sigma_tau_significance is not None
    ):
        return flex.bool(
            b.cc_one_half_sigma_tau_significance for b in merging_stats.bins
        ).reversed()
    elif merging_stats.overall.cc_one_half_significance is not None:
        return flex.bool(
            b.cc_one_half_significance for b in merging_stats.bins
        ).reversed()


def _get_cc_half_critical_values(merging_stats, cc_half_method):
    """Get the CC½ critical values from the input merging_stats object"""
    if (
        cc_half_method == "sigma_tau"
        and merging_stats.overall.cc_one_half_sigma_tau_critical_value is not None
    ):
        return flex.double(
            b.cc_one_half_sigma_tau_critical_value for b in merging_stats.bins
        ).reversed()
    elif merging_stats.overall.cc_one_half_critical_value is not None:
        critical = [
            b.cc_one_half_critical_value
            if b.cc_one_half_critical_value is not None
            else 0.0
            for b in merging_stats.bins
        ]
        return flex.double(critical).reversed()


def resolution_cc_half(
    merging_stats, limit, cc_half_method="half_dataset", model=tanh_fit
):
    """Estimate a resolution limit based on CC½

    The function defined by `model` will be fit to the CC½ values that have been
    pre-calculated by the `merging_stats` object. The estimated resolution limit is
    chosen as the `d_star_sq` value at which the fitted function equals `limit`.

    Args:
        merging_stats (iotbx.merging_statistics.dataset_statistics): Pre-calculated
            merging statistics object
        cc_half_method (str): The method for calculating CC½. Either "half_dataset" or
            "sigma_tau" (See Assmann et al., J. Appl. Cryst. (2016). 49, 1021–1028).
        model: The function to fit to the selected metric. Must be callable, taking as
            input x (d_star_sq) and y (the metric to be fitted) values, returning the
            fitted y(x) values. Default is `tanh_fit`.
        limit (float): The resolution limit criterion.

    Returns: The estimated resolution limit in units of Å^-1
    """
    sel = _get_cc_half_significance(merging_stats, cc_half_method)
    metric = "cc_one_half_sigma_tau" if cc_half_method == "sigma_tau" else "cc_one_half"
    result = resolution_fit_from_merging_stats(
        merging_stats, metric, model, limit, sel=sel
    )
    critical_values = _get_cc_half_critical_values(merging_stats, cc_half_method)
    if critical_values:
        result = result._replace(critical_values=critical_values.select(sel))
    return result


def interpolate_value(x, y, t):
    """Find the value of x: y(x) = t."""

    if t > max(y) or t < min(y):
        raise RuntimeError(f"t outside of [{min(y):f}, {max(y):f}]")

    for j in range(1, len(x)):
        x0 = x[j - 1]
        y0 = y[j - 1]

        x1 = x[j]
        y1 = y[j]

        if (y0 - t) * (y1 - t) < 0:
            return x0 + (t - y0) * (x1 - x0) / (y1 - y0)


def miller_array_from_mtz(unmerged_mtz, anomalous=False, labels=None):
    mtz_object = iotbx.mtz.object(file_name=unmerged_mtz)
    miller_arrays = mtz_object.as_miller_arrays(
        merge_equivalents=False, anomalous=anomalous
    )
    i_obs = None
    batches = None
    all_i_obs = []
    for array in miller_arrays:
        labels = array.info().label_string()
        if array.is_xray_intensity_array():
            all_i_obs.append(array)
        if labels == "BATCH":
            assert batches is None
            batches = array
    if i_obs is None:
        if len(all_i_obs) == 0:
            raise Sorry("No intensities found")
        elif len(all_i_obs) > 1:
            if labels is not None:
                lab_tab = label_table(all_i_obs)
                i_obs = lab_tab.select_array(
                    label=labels[0], command_line_switch="labels"
                )
            if i_obs is None:
                raise Sorry(
                    "Multiple intensity arrays - please specify one:\n%s"
                    % "\n".join(
                        ["  labels=%s" % a.info().label_string() for a in all_i_obs]
                    )
                )
        else:
            i_obs = all_i_obs[0]
    # need original miller indices otherwise we don't get correct anomalous
    # merging statistics
    if "M_ISYM" in mtz_object.column_labels():
        indices = mtz_object.extract_original_index_miller_indices()
        i_obs = i_obs.customized_copy(indices=indices, info=i_obs.info())
    return i_obs, batches


phil_str = """
  rmerge = None
    .type = float(value_min=0)
    .help = "Maximum value of Rmerge in the outer resolution shell"
    .short_caption = "Outer shell Rmerge"
    .expert_level = 1
  completeness = None
    .type = float(value_min=0)
    .help = "Minimum completeness in the outer resolution shell"
    .short_caption = "Outer shell completeness"
    .expert_level = 1
  cc_ref = 0.1
    .type = float(value_min=0)
    .help = "Minimum value of CC vs reference data set in the outer resolution shell"
    .short_caption = "Outer shell CCref"
    .expert_level = 1
  cc_half = 0.3
    .type = float(value_min=0)
    .help = "Minimum value of CC½ in the outer resolution shell"
    .short_caption = "Outer shell CC½"
    .expert_level = 1
  cc_half_method = *half_dataset sigma_tau
    .type = choice
    .short_caption = "CC½ method"
  cc_half_significance_level = 0.1
    .type = float(value_min=0, value_max=1)
    .expert_level = 1
    .short_caption = "CC½ significance level"
  cc_half_fit = polynomial *tanh
    .type = choice
    .expert_level = 1
    .short_caption = "CC½ fit"
  isigma = None
    .type = float(value_min=0)
    .help = "Minimum value of the unmerged <I/sigI> in the outer resolution shell"
    .short_caption = "Outer shell unmerged <I/sigI>"
    .expert_level = 1
  misigma = None
    .type = float(value_min=0)
    .help = "Minimum value of the merged <I/sigI> in the outer resolution shell"
    .short_caption = "Outer shell merged <I/sigI>"
    .expert_level = 1
  i_mean_over_sigma_mean = None
    .type = float(value_min=0)
    .help = "Minimum value of the unmerged <I>/<sigI> in the outer resolution shell"
    .short_caption = "Outer shell unmerged <I>/<sigI>"
    .expert_level = 2
  nbins = 100
    .type = int
    .help = "Maximum number of resolution bins to use for estimation of resolution limit."
    .short_caption = "Number of resolution bins."
    .expert_level = 1
  reflections_per_bin = 10
    .type = int
    .help = "Minimum number of reflections per bin."
    .short_caption = "Minimum number of reflections per bin"
  binning_method = *counting_sorted volume
    .type = choice
    .help = "Use equal-volume bins or bins with approximately equal numbers of reflections per bin."
    .short_caption = "Equal-volume or equal #ref binning."
    .expert_level = 1
  anomalous = False
    .type = bool
    .help = "Keep anomalous pairs separate in merging statistics"
    .short_caption = "Anomalous"
    .expert_level = 1
  labels = None
    .type = strings
    .short_caption = "Labels"
  space_group = None
    .type = space_group
    .expert_level = 1
    .short_caption = "Space group"
  reference = None
    .type = path
    .short_caption = "Reference"
  emax = 4
    .type = float(value_min = 0)
    .help = "Reject reflections with normalised intensities E^2 > emax^2"
    .short_caption = "Maximum normalised intensity"
"""


phil_defaults = iotbx.phil.parse(
    """
resolution {
%s
  batch_range = None
    .type = ints(size=2, value_min=0)
}
"""
    % phil_str
)


def plot_result(metric, result):
    if metric == metrics.CC_HALF:
        return plots.cc_half_plot(
            result.d_star_sq,
            result.y_obs,
            cc_half_critical_values=result.critical_values,
            cc_half_fit=result.y_fit,
            d_min=result.d_min,
        )
    else:
        d = {
            metrics.MISIGMA: "Merged <I/σ(I)>",
            metrics.ISIGMA: "Unmerged <I/σ(I)>",
            metrics.I_MEAN_OVER_SIGMA_MEAN: "&lt;I&gt;/<σ(I)>",
            metrics.RMERGE: "R<sub>merge</sub> ",
            metrics.COMPLETENESS: "Completeness",
        }
        d_star_sq_tickvals, d_star_sq_ticktext = plots.d_star_sq_to_d_ticks(
            result.d_star_sq, 5
        )
        return {
            "data": [
                {
                    "x": list(result.d_star_sq),  # d_star_sq
                    "y": list(result.y_obs),
                    "type": "scatter",
                    "name": "y_obs",
                },
                (
                    {
                        "x": list(result.d_star_sq),
                        "y": list(result.y_fit),
                        "type": "scatter",
                        "name": "y_fit",
                        "line": {"color": "rgb(47, 79, 79)"},
                    }
                    if result.y_fit
                    else {}
                ),
                (
                    {
                        "x": [uctbx.d_as_d_star_sq(result.d_min)] * 2,
                        "y": [
                            0,
                            max(
                                1,
                                flex.max(result.y_obs),
                                flex.max(result.y_fit) if result.y_fit else 0,
                            ),
                        ],
                        "type": "scatter",
                        "name": f"d_min = {result.d_min:.2f} Å",
                        "mode": "lines",
                        "line": {"color": "rgb(169, 169, 169)", "dash": "dot"},
                    }
                    if result.d_min
                    else {}
                ),
            ],
            "layout": {
                "title": f"{d.get(metric)} vs. resolution",
                "xaxis": {
                    "title": "Resolution (Å)",
                    "tickvals": d_star_sq_tickvals,
                    "ticktext": d_star_sq_ticktext,
                },
                "yaxis": {"title": d.get(metric), "rangemode": "tozero"},
            },
        }


class ResolutionResult(typing.NamedTuple):
    d_star_sq: flex.double
    y_obs: flex.double
    y_fit: flex.double
    d_min: float
    critical_values: flex.double = None


class Resolutionizer:
    """A class to calculate things from merging reflections."""

    def __init__(self, i_obs, params, batches=None, reference=None):

        self._params = params
        self._reference = reference

        if self._reference is not None:
            self._reference = self._reference.merge_equivalents(
                use_internal_variance=False
            ).array()

        i_obs = i_obs.customized_copy(
            anomalous_flag=params.anomalous, info=i_obs.info()
        )

        if self._params.batch_range is not None and batches is not None:
            batch_min, batch_max = self._params.batch_range
            assert batches is not None
            sel = (batches.data() >= batch_min) & (batches.data() <= batch_max)
            i_obs = i_obs.select(sel).set_info(i_obs.info())

        if self._params.space_group is not None:
            i_obs = i_obs.customized_copy(
                space_group_info=self._params.space_group, info=i_obs.info()
            )

        if self._params.emax:
            normalised = quasi_normalisation(i_obs)
            e2_cutoff = self._params.emax**2
            sel = normalised.data() < e2_cutoff
            logger.info(
                f"Removing {sel.count(False)} Wilson outliers with E^2 >= {e2_cutoff}"
            )
            i_obs = i_obs.select(sel)

        self._intensities = i_obs

        self._merging_statistics = iotbx.merging_statistics.dataset_statistics(
            i_obs=i_obs,
            n_bins=self._params.nbins,
            reflections_per_bin=self._params.reflections_per_bin,
            cc_one_half_significance_level=self._params.cc_half_significance_level,
            cc_one_half_method=self._params.cc_half_method,
            binning_method=self._params.binning_method,
            anomalous=params.anomalous,
            use_internal_variance=False,
            eliminate_sys_absent=False,
            assert_is_not_unique_set_under_symmetry=False,
        )

    @classmethod
    def from_unmerged_mtz(cls, scaled_unmerged, params):
        """Construct the resolutionizer from an mtz file."""

        i_obs, batches = miller_array_from_mtz(
            scaled_unmerged, anomalous=params.anomalous, labels=params.labels
        )
        if params.reference is not None:
            reference, _ = miller_array_from_mtz(
                params.reference, anomalous=params.anomalous, labels=params.labels
            )
        else:
            reference = None

        return cls(i_obs, params, batches=batches, reference=reference)

    @classmethod
    def from_reflections_and_experiments(cls, reflection_tables, experiments, params):
        """Construct the resolutionizer from native dials datatypes."""
        # add some assertions about data

        # do batch assignment (same functions as in dials.export)
        offsets = calculate_batch_offsets(experiments)
        reflection_tables = assign_batches_to_reflections(reflection_tables, offsets)
        batches = flex.int()
        intensities = flex.double()
        indices = flex.miller_index()
        variances = flex.double()
        for table in reflection_tables:
            if "intensity.scale.value" in table:
                table = filter_reflection_table(
                    table, ["scale"], partiality_threshold=0.4
                )
                intensities.extend(table["intensity.scale.value"])
                variances.extend(table["intensity.scale.variance"])
            else:
                table = filter_reflection_table(
                    table, ["profile"], partiality_threshold=0.4
                )
                intensities.extend(table["intensity.prf.value"])
                variances.extend(table["intensity.prf.variance"])
            indices.extend(table["miller_index"])
            batches.extend(table["batch"])

        crystal_symmetry = miller.crystal.symmetry(
            unit_cell=determine_best_unit_cell(experiments),
            space_group=experiments[0].crystal.get_space_group(),
            assert_is_compatible_unit_cell=False,
        )
        miller_set = miller.set(crystal_symmetry, indices, anomalous_flag=False)
        i_obs = miller.array(miller_set, data=intensities, sigmas=flex.sqrt(variances))
        i_obs.set_observation_type_xray_intensity()
        i_obs.set_info(miller.array_info(source="DIALS", source_type="refl"))

        ms = i_obs.customized_copy()
        batch_array = miller.array(ms, data=batches)

        if params.reference is not None:
            reference, _ = miller_array_from_mtz(
                params.reference, anomalous=params.anomalous, labels=params.labels
            )
        else:
            reference = None

        return cls(i_obs, params, batches=batch_array, reference=reference)

    def resolution(self, metric, limit=None):
        if metric == metrics.CC_HALF:
            return resolution_cc_half(
                self._merging_statistics,
                limit,
                cc_half_method=self._params.cc_half_method,
                model=tanh_fit
                if self._params.cc_half_fit == "tanh"
                else polynomial_fit,
            )
        elif metric == metrics.CC_REF:
            return self._resolution_cc_ref(limit=self._params.cc_ref)
        else:
            model = {
                metrics.RMERGE: log_inv_fit,
                metrics.COMPLETENESS: polynomial_fit,
                metrics.ISIGMA: log_fit,
                metrics.MISIGMA: log_fit,
                metrics.I_MEAN_OVER_SIGMA_MEAN: log_fit,
            }[metric]
            return resolution_fit_from_merging_stats(
                self._merging_statistics, metric.value, model, limit
            )

    def resolution_auto(self):
        """Compute resolution limits based on the current self._params set."""

        metric_to_output = {
            metrics.ISIGMA: "I/sig",
            metrics.MISIGMA: "Mn(I/sig)",
            metrics.I_MEAN_OVER_SIGMA_MEAN: "Mn(I)/Mn(sig)",
        }

        plot_d = {}

        for metric in metrics:
            name = metric.name.lower()
            limit = getattr(self._params, name)
            if metric == metrics.CC_REF and not self._reference:
                limit = None
            if limit:
                try:
                    result = self.resolution(metric, limit=limit)
                except RuntimeError as e:
                    logger.info(f"Resolution fit against {name} failed: {e}")
                    continue
                pretty_name = metric_to_output.get(metric, name)
                if result.d_min:
                    logger.info(
                        f"Resolution {pretty_name}:{result.d_min:{18 - len(pretty_name)}.2f}"
                    )
                plot_d[name] = plot_result(metric, result)
        return plot_d

    def _resolution_cc_ref(self, limit=None):
        """Compute a resolution limit where cc_ref < 0.5 (limit if
        set) or the full extent of the data."""

        if limit is None:
            limit = self._params.cc_ref

        intensities = self._intensities.merge_equivalents(
            use_internal_variance=False
        ).array()
        cc_s = flex.double()
        for b in self._merging_statistics.bins:
            cc = intensities.resolution_filter(
                d_min=b.d_min, d_max=b.d_max
            ).correlation(
                self._reference.resolution_filter(d_min=b.d_min, d_max=b.d_max),
                assert_is_similar_symmetry=False,
            )
            cc_s.append(cc.coefficient())
        cc_s = cc_s.reversed()

        fit = tanh_fit if self._params.cc_half_fit == "tanh" else polynomial_fit
        d_star_sq = flex.double(
            1 / b.d_min**2 for b in self._merging_statistics.bins
        ).reversed()

        return resolution_fit(d_star_sq, cc_s, fit, limit)



"""
Contains implementation interface for finding spots on one or many images
"""

from __future__ import annotations

import logging
import math
import pickle
from typing import Iterable, Tuple

import libtbx
from dxtbx.format.image import ImageBool
from dxtbx.imageset import ImageSequence, ImageSet
from dxtbx.model import ExperimentList

from dials.array_family import flex
from dials.model.data import PixelList, PixelListLabeller
from dials.util import Sorry, log
from dials.util.log import rehandle_cached_records
from dials.util.mp import available_cores, batch_multi_node_parallel_map

logger = logging.getLogger(__name__)


class ExtractPixelsFromImage:
    """
    A class to extract pixels from a single image
    """

    def __init__(
        self,
        imageset,
        threshold_function,
        mask,
        region_of_interest,
        max_strong_pixel_fraction,
        compute_mean_background,
    ):
        """
        Initialise the class

        :param imageset: The imageset to extract from
        :param threshold_function: The function to threshold with
        :param mask: The image mask
        :param region_of_interest: A region of interest to process
        :param max_strong_pixel_fraction: The maximum fraction of pixels allowed
        """
        self.threshold_function = threshold_function
        self.imageset = imageset
        self.mask = mask
        self.region_of_interest = region_of_interest
        self.max_strong_pixel_fraction = max_strong_pixel_fraction
        self.compute_mean_background = compute_mean_background
        if self.mask is not None:
            detector = self.imageset.get_detector()
            assert len(self.mask) == len(detector)

    def __call__(self, index):
        """
        Extract strong pixels from an image

        :param index: The index of the image
        """
        # Get the frame number
        if isinstance(self.imageset, ImageSequence):
            frame = self.imageset.get_array_range()[0] + index
        else:
            ind = self.imageset.indices()
            if len(ind) > 1:
                assert all(i1 + 1 == i2 for i1, i2 in zip(ind[0:-1], ind[1:-1]))
            frame = ind[index]

        # Create the list of pixel lists
        pixel_list = []

        # Get the image and mask
        image = self.imageset.get_corrected_data(index)
        mask = self.imageset.get_mask(index)

        # Set the mask
        if self.mask is not None:
            assert len(self.mask) == len(mask)
            mask = tuple(m1 & m2 for m1, m2 in zip(mask, self.mask))

        logger.debug(
            "Number of masked pixels for image %i: %i",
            index,
            sum(m.count(False) for m in mask),
        )

        # Add the images to the pixel lists
        num_strong = 0
        average_background = 0
        for i_panel, (im, mk) in enumerate(zip(image, mask)):
            if self.region_of_interest is not None:
                x0, x1, y0, y1 = self.region_of_interest
                height, width = im.all()
                assert x0 < x1, "x0 < x1"
                assert y0 < y1, "y0 < y1"
                assert x0 >= 0, "x0 >= 0"
                assert y0 >= 0, "y0 >= 0"
                assert x1 <= width, "x1 <= width"
                assert y1 <= height, "y1 <= height"
                im_roi = im[y0:y1, x0:x1]
                mk_roi = mk[y0:y1, x0:x1]
                tm_roi = self.threshold_function.compute_threshold(
                    im_roi,
                    mk_roi,
                    imageset=self.imageset,
                    i_panel=i_panel,
                    region_of_interest=self.region_of_interest,
                )
                threshold_mask = flex.bool(im.accessor(), False)
                threshold_mask[y0:y1, x0:x1] = tm_roi
            else:
                threshold_mask = self.threshold_function.compute_threshold(
                    im, mk, imageset=self.imageset, i_panel=i_panel
                )

            # Add the pixel list
            plist = PixelList(frame, im, threshold_mask)
            pixel_list.append(plist)

            # Get average background
            if self.compute_mean_background:
                background = im.as_1d().select((mk & ~threshold_mask).as_1d())
                average_background += flex.mean(background)

            # Add to the spot count
            num_strong += len(plist)

        # Make average background
        average_background /= len(image)

        # Check total number of strong pixels
        if self.max_strong_pixel_fraction < 1:
            num_image = 0
            for im in image:
                num_image += len(im)
            max_strong = int(math.ceil(self.max_strong_pixel_fraction * num_image))
            if num_strong > max_strong:
                raise RuntimeError(
                    f"""
          The number of strong pixels found ({num_strong}) is greater than the
          maximum allowed ({max_strong}). Try changing spot finding parameters
        """
                )

        # Print some info
        if self.compute_mean_background:
            logger.info(
                "Found %d strong pixels on image %d with average background %f",
                num_strong,
                frame + 1,
                average_background,
            )
        else:
            logger.info("Found %d strong pixels on image %d", num_strong, frame + 1)

        # Return the result
        return pixel_list


class ExtractPixelsFromImage2DNoShoeboxes(ExtractPixelsFromImage):
    """
    A class to extract pixels from a single image
    """

    def __init__(
        self,
        imageset,
        threshold_function,
        mask,
        region_of_interest,
        max_strong_pixel_fraction,
        compute_mean_background,
        min_spot_size,
        max_spot_size,
        filter_spots,
    ):
        """
        Initialise the class

        :param imageset: The imageset to extract from
        :param threshold_function: The function to threshold with
        :param mask: The image mask
        :param region_of_interest: A region of interest to process
        :param max_strong_pixel_fraction: The maximum fraction of pixels allowed
        """
        super().__init__(
            imageset,
            threshold_function,
            mask,
            region_of_interest,
            max_strong_pixel_fraction,
            compute_mean_background,
        )

        # Save some stuff
        self.min_spot_size = min_spot_size
        self.max_spot_size = max_spot_size
        self.filter_spots = filter_spots

    def __call__(self, index):
        """
        Extract strong pixels from an image

        :param index: The index of the image
        """
        # Initialise the pixel labeller
        num_panels = len(self.imageset.get_detector())
        pixel_labeller = [PixelListLabeller() for p in range(num_panels)]

        # Call the super function
        result = super().__call__(index)

        # Add pixel lists to the labeller
        assert len(pixel_labeller) == len(result), "Inconsistent size"
        for plabeller, plist in zip(pixel_labeller, result):
            plabeller.add(plist)

        # Create shoeboxes from pixel list
        reflections, _ = pixel_list_to_reflection_table(
            self.imageset,
            pixel_labeller,
            filter_spots=self.filter_spots,
            min_spot_size=self.min_spot_size,
            max_spot_size=self.max_spot_size,
            write_hot_pixel_mask=False,
        )

        # Delete the shoeboxes
        del reflections["shoeboxes"]

        # Return the reflections
        return [reflections]


class ExtractSpotsParallelTask:
    """
    Execute the spot finder task in parallel

    We need this external class so that we can pickle it for cluster jobs
    """

    def __init__(self, function):
        """
        Initialise with the function to call
        """
        self.function = function

    def __call__(self, task):
        """
        Call the function with th task and save the IO
        """
        log.config_simple_cached()
        result = self.function(task)
        handlers = logging.getLogger("dials").handlers
        assert len(handlers) == 1, "Invalid number of logging handlers"
        return result, handlers[0].records


def pixel_list_to_shoeboxes(
    imageset: ImageSet,
    pixel_labeller: Iterable[PixelListLabeller],
    min_spot_size: int,
    max_spot_size: int,
    write_hot_pixel_mask: bool,
) -> Tuple[flex.shoebox, Tuple[flex.size_t, ...]]:
    """Convert a pixel list to shoeboxes"""
    # Extract the pixel lists into a list of reflections
    shoeboxes = flex.shoebox()
    spotsizes = flex.size_t()
    hotpixels = tuple(flex.size_t() for i in range(len(imageset.get_detector())))
    if isinstance(imageset, ImageSequence):
        twod = imageset.get_scan().is_still()
    else:
        twod = True
    for i, (p, hp) in enumerate(zip(pixel_labeller, hotpixels)):
        if p.num_pixels() > 0:
            creator = flex.PixelListShoeboxCreator(
                p,
                i,  # panel
                0,  # zrange
                twod,  # twod
                min_spot_size,  # min_pixels
                max_spot_size,  # max_pixels
                write_hot_pixel_mask,
            )
            shoeboxes.extend(creator.result())
            spotsizes.extend(creator.spot_size())
            hp.extend(creator.hot_pixels())
    logger.info("\nExtracted %d spots", len(shoeboxes))

    # Get the unallocated spots and print some info
    selection = shoeboxes.is_allocated()
    shoeboxes = shoeboxes.select(selection)
    ntoosmall = (spotsizes < min_spot_size).count(True)
    ntoolarge = (spotsizes > max_spot_size).count(True)
    assert ntoosmall + ntoolarge == selection.count(False)
    logger.info("Removed %d spots with size < %d pixels", ntoosmall, min_spot_size)
    logger.info("Removed %d spots with size > %d pixels", ntoolarge, max_spot_size)

    # Return the shoeboxes
    return shoeboxes, hotpixels


def shoeboxes_to_reflection_table(
    imageset: ImageSet, shoeboxes: flex.shoebox, filter_spots
) -> flex.reflection_table:
    """Filter shoeboxes and create reflection table"""
    # Calculate the spot centroids
    centroid = shoeboxes.centroid_valid()
    logger.info("Calculated %d spot centroids", len(shoeboxes))

    # Calculate the spot intensities
    intensity = shoeboxes.summed_intensity()
    logger.info("Calculated %d spot intensities", len(shoeboxes))

    # Create the observations
    observed = flex.observation(shoeboxes.panels(), centroid, intensity)

    # Filter the reflections and select only the desired spots
    flags = filter_spots(
        None, sweep=imageset, observations=observed, shoeboxes=shoeboxes
    )
    observed = observed.select(flags)
    shoeboxes = shoeboxes.select(flags)

    # Return as a reflection list
    return flex.reflection_table(observed, shoeboxes)


def pixel_list_to_reflection_table(
    imageset: ImageSet,
    pixel_labeller: Iterable[PixelListLabeller],
    filter_spots,
    min_spot_size: int,
    max_spot_size: int,
    write_hot_pixel_mask: bool,
) -> Tuple[flex.shoebox, Tuple[flex.size_t, ...]]:
    """Convert pixel list to reflection table"""
    shoeboxes, hot_pixels = pixel_list_to_shoeboxes(
        imageset,
        pixel_labeller,
        min_spot_size=min_spot_size,
        max_spot_size=max_spot_size,
        write_hot_pixel_mask=write_hot_pixel_mask,
    )
    # Setup the reflection table converter
    return (
        shoeboxes_to_reflection_table(imageset, shoeboxes, filter_spots=filter_spots),
        hot_pixels,
    )


class ExtractSpots:
    """
    Class to find spots in an image and extract them into shoeboxes.
    """

    def __init__(
        self,
        threshold_function=None,
        mask=None,
        region_of_interest=None,
        max_strong_pixel_fraction=0.1,
        compute_mean_background=False,
        mp_method=None,
        mp_nproc=1,
        mp_njobs=1,
        mp_chunksize=1,
        min_spot_size=1,
        max_spot_size=20,
        filter_spots=None,
        no_shoeboxes_2d=False,
        min_chunksize=50,
        write_hot_pixel_mask=False,
    ):
        """
        Initialise the class with the strategy

        :param threshold_function: The image thresholding strategy
        :param mask: The mask to use
        :param mp_method: The multi processing method
        :param nproc: The number of processors
        :param max_strong_pixel_fraction: The maximum number of strong pixels
        """
        # Set the required strategies
        self.threshold_function = threshold_function
        self.mask = mask
        self.mp_method = mp_method
        self.mp_chunksize = mp_chunksize
        self.mp_nproc = mp_nproc
        self.mp_njobs = mp_njobs
        self.max_strong_pixel_fraction = max_strong_pixel_fraction
        self.compute_mean_background = compute_mean_background
        self.region_of_interest = region_of_interest
        self.min_spot_size = min_spot_size
        self.max_spot_size = max_spot_size
        self.filter_spots = filter_spots
        self.no_shoeboxes_2d = no_shoeboxes_2d
        self.min_chunksize = min_chunksize
        self.write_hot_pixel_mask = write_hot_pixel_mask

    def __call__(self, imageset):
        """
        Find the spots in the imageset

        :param imageset: The imageset to process
        :return: The list of spot shoeboxes
        """
        if not self.no_shoeboxes_2d:
            return self._find_spots(imageset)
        else:
            return self._find_spots_2d_no_shoeboxes(imageset)

    def _compute_chunksize(self, nimg, nproc, min_chunksize):
        """
        Compute the chunk size for a given number of images and processes
        """
        chunksize = int(math.ceil(nimg / nproc))
        remainder = nimg % (chunksize * nproc)
        test_chunksize = chunksize - 1
        while test_chunksize >= min_chunksize:
            test_remainder = nimg % (test_chunksize * nproc)
            if test_remainder <= remainder:
                chunksize = test_chunksize
                remainder = test_remainder
            test_chunksize -= 1
        return chunksize

    def _find_spots(self, imageset):
        """
        Find the spots in the imageset

        :param imageset: The imageset to process
        :return: The list of spot shoeboxes
        """
        # Change the number of processors if necessary
        mp_nproc = self.mp_nproc
        mp_njobs = self.mp_njobs
        if mp_nproc is libtbx.Auto:
            mp_nproc = available_cores()
            logger.info(f"Setting nproc={mp_nproc}")
        if mp_nproc * mp_njobs > len(imageset):
            mp_nproc = min(mp_nproc, len(imageset))
            mp_njobs = int(math.ceil(len(imageset) / mp_nproc))

        mp_method = self.mp_method
        mp_chunksize = self.mp_chunksize

        if mp_chunksize is libtbx.Auto:
            mp_chunksize = self._compute_chunksize(
                len(imageset), mp_njobs * mp_nproc, self.min_chunksize
            )
            logger.info("Setting chunksize=%i", mp_chunksize)

        len_by_nproc = int(math.floor(len(imageset) / (mp_njobs * mp_nproc)))
        if mp_chunksize > len_by_nproc:
            mp_chunksize = len_by_nproc
        if mp_chunksize == 0:
            mp_chunksize = 1
        assert mp_nproc > 0, "Invalid number of processors"
        assert mp_njobs > 0, "Invalid number of jobs"
        assert mp_njobs == 1 or mp_method is not None, "Invalid cluster method"
        assert mp_chunksize > 0, "Invalid chunk size"

        # The extract pixels function
        function = ExtractPixelsFromImage(
            imageset=imageset,
            threshold_function=self.threshold_function,
            mask=self.mask,
            max_strong_pixel_fraction=self.max_strong_pixel_fraction,
            compute_mean_background=self.compute_mean_background,
            region_of_interest=self.region_of_interest,
        )

        # The indices to iterate over
        indices = list(range(len(imageset)))

        # Initialise the pixel labeller
        num_panels = len(imageset.get_detector())
        pixel_labeller = [PixelListLabeller() for p in range(num_panels)]

        # Do the processing
        logger.info("Extracting strong pixels from images")
        if mp_njobs > 1:
            logger.info(
                " Using %s with %d parallel job(s) and %d processes per node\n",
                mp_method,
                mp_njobs,
                mp_nproc,
            )
        else:
            logger.info(" Using multiprocessing with %d parallel job(s)\n", mp_nproc)
        if mp_nproc > 1 or mp_njobs > 1:

            def process_output(result):
                rehandle_cached_records(result[1])
                assert len(pixel_labeller) == len(result[0]), "Inconsistent size"
                for plabeller, plist in zip(pixel_labeller, result[0]):
                    plabeller.add(plist)

            batch_multi_node_parallel_map(
                func=ExtractSpotsParallelTask(function),
                iterable=indices,
                nproc=mp_nproc,
                njobs=mp_njobs,
                cluster_method=mp_method,
                chunksize=mp_chunksize,
                callback=process_output,
            )
        else:
            for task in indices:
                result = function(task)
                assert len(pixel_labeller) == len(result), "Inconsistent size"
                for plabeller, plist in zip(pixel_labeller, result):
                    plabeller.add(plist)
                result.clear()

        # Create shoeboxes from pixel list
        return pixel_list_to_reflection_table(
            imageset,
            pixel_labeller,
            filter_spots=self.filter_spots,
            min_spot_size=self.min_spot_size,
            max_spot_size=self.max_spot_size,
            write_hot_pixel_mask=self.write_hot_pixel_mask,
        )

    def _find_spots_2d_no_shoeboxes(self, imageset):
        """
        Find the spots in the imageset

        :param imageset: The imageset to process
        :return: The list of spot shoeboxes
        """
        # Change the number of processors if necessary
        mp_nproc = self.mp_nproc
        mp_njobs = self.mp_njobs
        if mp_nproc * mp_njobs > len(imageset):
            mp_nproc = min(mp_nproc, len(imageset))
            mp_njobs = int(math.ceil(len(imageset) / mp_nproc))

        mp_method = self.mp_method
        mp_chunksize = self.mp_chunksize

        if mp_chunksize == libtbx.Auto:
            mp_chunksize = self._compute_chunksize(
                len(imageset), mp_njobs * mp_nproc, self.min_chunksize
            )
            logger.info("Setting chunksize=%i", mp_chunksize)

        len_by_nproc = int(math.floor(len(imageset) / (mp_njobs * mp_nproc)))
        if mp_chunksize > len_by_nproc:
            mp_chunksize = len_by_nproc
        assert mp_nproc > 0, "Invalid number of processors"
        assert mp_njobs > 0, "Invalid number of jobs"
        assert mp_njobs == 1 or mp_method is not None, "Invalid cluster method"
        assert mp_chunksize > 0, "Invalid chunk size"

        # The extract pixels function
        function = ExtractPixelsFromImage2DNoShoeboxes(
            imageset=imageset,
            threshold_function=self.threshold_function,
            mask=self.mask,
            max_strong_pixel_fraction=self.max_strong_pixel_fraction,
            compute_mean_background=self.compute_mean_background,
            region_of_interest=self.region_of_interest,
            min_spot_size=self.min_spot_size,
            max_spot_size=self.max_spot_size,
            filter_spots=self.filter_spots,
        )

        # The indices to iterate over
        indices = list(range(len(imageset)))

        # The resulting reflections
        reflections = flex.reflection_table()

        # Do the processing
        logger.info("Extracting strong spots from images")
        if mp_njobs > 1:
            logger.info(
                " Using %s with %d parallel job(s) and %d processes per node\n",
                mp_method,
                mp_njobs,
                mp_nproc,
            )
        else:
            logger.info(" Using multiprocessing with %d parallel job(s)\n", mp_nproc)
        if mp_nproc > 1 or mp_njobs > 1:

            def process_output(result):
                for message in result[1]:
                    logger.log(message.levelno, message.msg)
                reflections.extend(result[0][0])
                result[0][0] = None

            batch_multi_node_parallel_map(
                func=ExtractSpotsParallelTask(function),
                iterable=indices,
                nproc=mp_nproc,
                njobs=mp_njobs,
                cluster_method=mp_method,
                chunksize=mp_chunksize,
                callback=process_output,
            )
        else:
            for task in indices:
                reflections.extend(function(task)[0])

        # Return the reflections
        return reflections, None


class SpotFinder:
    """
    A class to do spot finding and filtering.
    """

    def __init__(
        self,
        threshold_function=None,
        mask=None,
        region_of_interest=None,
        max_strong_pixel_fraction=0.1,
        compute_mean_background=False,
        mp_method=None,
        mp_nproc=1,
        mp_njobs=1,
        mp_chunksize=1,
        mask_generator=None,
        filter_spots=None,
        scan_range=None,
        write_hot_mask=True,
        hot_mask_prefix="hot_mask",
        min_spot_size=1,
        max_spot_size=20,
        no_shoeboxes_2d=False,
        min_chunksize=50,
        is_stills=False,
    ):
        """
        Initialise the class.

        :param find_spots: The spot finding algorithm
        :param filter_spots: The spot filtering algorithm
        :param scan_range: The scan range to find spots over
        :param is_stills:   [ADVANCED] Force still-handling of experiment
                            ID remapping for dials.stills_process.
        """

        # Set the filter and some other stuff
        self.threshold_function = threshold_function
        self.mask = mask
        self.region_of_interest = region_of_interest
        self.max_strong_pixel_fraction = max_strong_pixel_fraction
        self.compute_mean_background = compute_mean_background
        self.mask_generator = mask_generator
        self.filter_spots = filter_spots
        self.scan_range = scan_range
        self.write_hot_mask = write_hot_mask
        self.hot_mask_prefix = hot_mask_prefix
        self.min_spot_size = min_spot_size
        self.max_spot_size = max_spot_size
        self.mp_method = mp_method
        self.mp_chunksize = mp_chunksize
        self.mp_nproc = mp_nproc
        self.mp_njobs = mp_njobs
        self.no_shoeboxes_2d = no_shoeboxes_2d
        self.min_chunksize = min_chunksize
        self.is_stills = is_stills

    def find_spots(self, experiments: ExperimentList) -> flex.reflection_table:
        """
        Do spotfinding for a set of experiments.

        Args:
            experiments: The experiment list to process

        Returns:
            A new reflection table of found reflections
        """
        # Loop through all the experiments and get the unique imagesets
        imagesets = []
        for experiment in experiments:
            if experiment.imageset not in imagesets:
                imagesets.append(experiment.imageset)

        # Loop through all the imagesets and find the strong spots
        reflections = flex.reflection_table()

        for j, imageset in enumerate(imagesets):

            # Find the strong spots in the sequence
            logger.info(
                "-" * 80 + "\nFinding strong spots in imageset %d\n" + "-" * 80, j
            )
            table, hot_mask = self._find_spots_in_imageset(imageset)

            # Fix up the experiment ID's now
            table["id"] = flex.int(table.nrows(), -1)
            for i, experiment in enumerate(experiments):
                if experiment.imageset is not imageset:
                    continue
                if not self.is_stills and experiment.scan:
                    z0, z1 = experiment.scan.get_array_range()
                    z = table["xyzobs.px.value"].parts()[2]
                    table["id"].set_selected((z > z0) & (z < z1), i)
                    if experiment.identifier:
                        table.experiment_identifiers()[i] = experiment.identifier
                else:
                    table["id"] = flex.int(table.nrows(), j)
                    if experiment.identifier:
                        table.experiment_identifiers()[j] = experiment.identifier
            missed = table["id"] == -1
            assert missed.count(True) == 0, "Failed to remap {} experiment IDs".format(
                missed.count(True)
            )

            reflections.extend(table)
            # Write a hot pixel mask
            if self.write_hot_mask:
                if not imageset.external_lookup.mask.data.empty():
                    for m1, m2 in zip(hot_mask, imageset.external_lookup.mask.data):
                        m1 &= m2.data()
                    imageset.external_lookup.mask.data = ImageBool(hot_mask)
                else:
                    imageset.external_lookup.mask.data = ImageBool(hot_mask)
                imageset.external_lookup.mask.filename = "%s_%d.pickle" % (
                    self.hot_mask_prefix,
                    i,
                )

                # Write the hot mask
                with open(imageset.external_lookup.mask.filename, "wb") as outfile:
                    pickle.dump(hot_mask, outfile, protocol=pickle.HIGHEST_PROTOCOL)

        # Set the strong spot flag
        reflections.set_flags(
            flex.size_t_range(len(reflections)), reflections.flags.strong
        )

        # Check for overloads
        reflections.is_overloaded(experiments)

        # Return the reflections
        return reflections

    def _find_spots_in_imageset(self, imageset):
        """
        Do the spot finding.

        :param imageset: The imageset to process
        :return: The observed spots
        """
        # The input mask
        mask = self.mask_generator(imageset)
        if self.mask is not None:
            mask = tuple(m1 & m2 for m1, m2 in zip(mask, self.mask))

        # Set the spot finding algorithm
        extract_spots = ExtractSpots(
            threshold_function=self.threshold_function,
            mask=mask,
            region_of_interest=self.region_of_interest,
            max_strong_pixel_fraction=self.max_strong_pixel_fraction,
            compute_mean_background=self.compute_mean_background,
            mp_method=self.mp_method,
            mp_nproc=self.mp_nproc,
            mp_njobs=self.mp_njobs,
            mp_chunksize=self.mp_chunksize,
            min_spot_size=self.min_spot_size,
            max_spot_size=self.max_spot_size,
            filter_spots=self.filter_spots,
            no_shoeboxes_2d=self.no_shoeboxes_2d,
            min_chunksize=self.min_chunksize,
            write_hot_pixel_mask=self.write_hot_mask,
        )

        # Get the max scan range
        if isinstance(imageset, ImageSequence):
            max_scan_range = imageset.get_array_range()
        else:
            max_scan_range = (0, len(imageset))

        # Get list of scan ranges
        if not self.scan_range or self.scan_range[0] is None:
            scan_range = [(max_scan_range[0] + 1, max_scan_range[1])]
        else:
            scan_range = self.scan_range

        # Get spots from bits of scan
        hot_pixels = tuple(flex.size_t() for i in range(len(imageset.get_detector())))
        reflections = flex.reflection_table()
        for j0, j1 in scan_range:
            # Make sure we were asked to do something sensible
            if j1 < j0:
                raise Sorry("Scan range must be in ascending order")
            elif j0 < max_scan_range[0] or j1 > max_scan_range[1]:
                raise Sorry(
                    "Scan range must be within image range {}..{}".format(
                        max_scan_range[0] + 1, max_scan_range[1]
                    )
                )

            logger.info("\nFinding spots in image %s to %s...", j0, j1)
            j0 -= 1
            if len(imageset) == 1:
                r, h = extract_spots(imageset)
            else:
                r, h = extract_spots(imageset[j0:j1])
            reflections.extend(r)
            if h is not None:
                for h1, h2 in zip(hot_pixels, h):
                    h1.extend(h2)

        # Find hot pixels
        hot_mask = self._create_hot_mask(imageset, hot_pixels)

        # Return as a reflection list
        return reflections, hot_mask

    def _create_hot_mask(self, imageset, hot_pixels):
        """
        Find hot pixels in images
        """
        # Write the hot mask
        if self.write_hot_mask:

            # Create the hot pixel mask
            hot_mask = tuple(
                flex.bool(flex.grid(p.get_image_size()[::-1]), True)
                for p in imageset.get_detector()
            )
            num_hot = 0
            if hot_pixels:
                for hp, hm in zip(hot_pixels, hot_mask):
                    for i in range(len(hp)):
                        hm[hp[i]] = False
                    num_hot += len(hp)
            logger.info("Found %d possible hot pixel(s)", num_hot)

        else:
            hot_mask = None

        # Return the hot mask
        return hot_mask



"""
Split a sequence into a set of stills.

Example:

  dials.sequence_to_stills sequence.expt sequence.refl
"""


from __future__ import annotations

import logging

from dxtbx.model import MosaicCrystalSauter2014
from dxtbx.model.experiment_list import Experiment, ExperimentList
from libtbx.phil import parse
from scitbx import matrix

from dials.algorithms.refinement.prediction.managed_predictors import (
    ExperimentsPredictorFactory,
)
from dials.array_family import flex
from dials.model.data import Shoebox
from dials.util import show_mail_handle_errors
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

logger = logging.getLogger("dials.command_line.sequence_to_stills")

# The phil scope
phil_scope = parse(
    """
output {
  experiments = stills.expt
    .type = str
    .help = "Filename for the experimental models that have been converted to stills"
  reflections = stills.refl
    .type = str
    .help = "Filename for the reflection tables with split shoeboxes (3D to 2D)"
  domain_size_ang = None
    .type = float
    .help = "Override for domain size. If None, use the crystal's domain size, if"
            "available"
  half_mosaicity_deg = None
    .type = float
    .help = "Override for mosaic angle. If None, use the crystal's mosaic angle, if"
            "available"
}
max_scan_points = None
  .type = int
  .expert_level = 2
  .help = Limit number of scan points
"""
)


def sequence_to_stills(experiments, reflections, params):
    assert len(reflections) == 1
    reflections = reflections[0]

    new_experiments = ExperimentList()
    new_reflections = flex.reflection_table()

    # This is the subset needed to integrate
    for key in [
        "id",
        "imageset_id",
        "shoebox",
        "bbox",
        "intensity.sum.value",
        "intensity.sum.variance",
        "entering",
        "flags",
        "miller_index",
        "panel",
        "xyzobs.px.value",
        "xyzobs.px.variance",
    ]:
        if key in reflections:
            new_reflections[key] = type(reflections[key])()
        elif key == "imageset_id":
            assert len(experiments.imagesets()) == 1
            reflections["imageset_id"] = flex.int(len(reflections), 0)
            new_reflections["imageset_id"] = flex.int()
        elif key == "entering":
            reflections["entering"] = flex.bool(len(reflections), False)
            new_reflections["entering"] = flex.bool()
        else:
            raise RuntimeError(f"Expected key not found in reflection table: {key}")

    for expt_id, experiment in enumerate(experiments):
        # Get the goniometer setting matrix
        goniometer_setting_matrix = matrix.sqr(
            experiment.goniometer.get_setting_rotation()
        )
        goniometer_axis = matrix.col(experiment.goniometer.get_rotation_axis())
        step = experiment.scan.get_oscillation()[1]

        refls = reflections.select(reflections["id"] == expt_id)
        _, _, _, _, z1, z2 = refls["bbox"].parts()

        # Create an experiment for each scanpoint.
        # Note that a simplification of the use of scan-points here introduces a
        # (small) error. The scan-varying crystal model has 1 more scan-point
        # than the scan has images because scan-points are taken at the boundaries
        # between images, including the scan extrema. This code assumes that
        # the crystal model at the start of each image applies to the whole
        # image and ignores the final scan-point.
        start, stop = experiment.scan.get_array_range()
        for i_array in range(start, stop):
            if params.max_scan_points and i_array >= params.max_scan_points:
                break
            # Shift array position to scan-point index
            i_scan_point = i_array - start

            # The A matrix is the goniometer setting matrix for this scan point
            # times the scan varying A matrix at this scan point. Note, the
            # goniometer setting matrix for scan point zero will be the identity
            # matrix and represents the beginning of the oscillation.
            # For stills, the A matrix needs to be positioned in the midpoint of an
            # oscillation step. Hence, here the goniometer setting matrix is rotated
            # by a further half oscillation step.
            A = (
                goniometer_axis.axis_and_angle_as_r3_rotation_matrix(
                    angle=experiment.scan.get_angle_from_array_index(i_array)
                    + (step / 2),
                    deg=True,
                )
                * goniometer_setting_matrix
                * matrix.sqr(experiment.crystal.get_A_at_scan_point(i_array))
            )
            crystal = MosaicCrystalSauter2014(experiment.crystal)
            crystal.set_A(A)

            # Copy in mosaic parameters if available
            if params.output.domain_size_ang is None and hasattr(
                experiment.crystal, "get_domain_size_ang"
            ):
                crystal.set_domain_size_ang(experiment.crystal.get_domain_size_ang())
            elif params.output.domain_size_ang is not None:
                crystal.set_domain_size_ang(params.output.domain_size_ang)

            if params.output.half_mosaicity_deg is None and hasattr(
                experiment.crystal, "get_half_mosaicity_deg"
            ):
                crystal.set_half_mosaicity_deg(
                    experiment.crystal.get_half_mosaicity_deg()
                )
            elif params.output.half_mosaicity_deg is not None:
                crystal.set_half_mosaicity_deg(params.output.half_mosaicity_deg)

            new_experiment = Experiment(
                detector=experiment.detector,
                beam=experiment.beam,
                crystal=crystal,
                imageset=experiment.imageset.as_imageset()[
                    i_scan_point : i_scan_point + 1
                ],
            )
            new_experiments.append(new_experiment)

            # Each reflection in a 3D shoebox can be found on multiple images.
            # Slice the reflections such that any reflection on this scan point
            # is included with this image
            new_id = len(new_experiments) - 1
            subrefls = refls.select((i_scan_point >= z1) & (i_scan_point < z2))
            for refl in subrefls.rows():
                assert i_scan_point in range(*refl["bbox"][4:6])

                new_sb = Shoebox()
                start = i_scan_point - refl["bbox"][4]  # z1
                new_sb.data = refl["shoebox"].data[start : start + 1, :, :]
                new_sb.background = refl["shoebox"].background[start : start + 1, :, :]
                new_sb.mask = refl["shoebox"].mask[start : start + 1, :, :]
                intensity = new_sb.summed_intensity()
                new_sb.bbox = tuple(
                    list(refl["bbox"])[0:4] + [0, 1]
                )  # keep the original shoebox but reset the z values
                new_sb.panel = refl["panel"]
                new_refl = {}
                new_refl["id"] = new_refl["imageset_id"] = new_id
                new_refl["shoebox"] = new_sb
                new_refl["bbox"] = new_sb.bbox
                new_refl["intensity.sum.value"] = intensity.observed.value
                new_refl["intensity.sum.variance"] = intensity.observed.variance
                for key in ["entering", "flags", "miller_index", "panel"]:
                    new_refl[key] = refl[key]
                centroid = new_sb.centroid_foreground_minus_background()
                new_refl["xyzobs.px.value"] = centroid.px.position
                new_refl["xyzobs.px.variance"] = centroid.px.variance
                new_reflections.append({})
                for key in new_refl:
                    new_reflections[key][-1] = new_refl[key]

    # Re-predict using the reflection slices and the stills predictors
    ref_predictor = ExperimentsPredictorFactory.from_experiments(
        new_experiments, force_stills=new_experiments.all_stills()
    )
    new_reflections = ref_predictor(new_reflections)

    return (new_experiments, new_reflections)


@show_mail_handle_errors()
def run(args=None, phil=phil_scope):
    """
    Validate the arguments and load experiments/reflections for sequence_to_stills

    Arguments:
        args: The command line arguments to use. Defaults to sys.argv[1:]
        phil: The phil_scope. Defaults to the master phil_scope for this program
    """
    # The script usage
    usage = "usage: dials.sequence_to_stills [options] [param.phil] models.expt reflections.refl"

    # Create the parser
    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        epilog=__doc__,
    )
    params, options = parser.parse_args(args=args, show_diff_phil=True)

    # Try to load the models and data
    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        return

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    (new_experiments, new_reflections) = sequence_to_stills(
        experiments, reflections, params
    )
    # Write out the output experiments, reflections
    new_experiments.as_file(params.output.experiments)
    new_reflections.as_file(params.output.reflections)


if __name__ == "__main__":
    run()



from __future__ import annotations

import io
from unittest import mock

import pytest

import boost_adaptbx.boost.python

from dials.util.ext import ostream, streambuf

ext = boost_adaptbx.boost.python.import_ext("dials_util_streambuf_test_ext")


class io_test_case:
    phrase = b"Coding should be fun"
    #          01234567890123456789

    def run(self):
        m = streambuf.default_buffer_size
        for n in range(50, 0, -1):
            streambuf.default_buffer_size = n
            self.exercise_read_failure()
            self.exercise_write_failure()
            self.exercise_read()
            self.exercise_write()
            self.exercise_seek_and_read()
            self.exercise_partial_read()
            self.exercise_write_and_seek()
        streambuf.default_buffer_size = m

    def exercise_read(self):
        self.create_file_object(mode="rb")
        words = ext.read_word(streambuf(self.file_object))
        assert words == b"Coding, should, be, fun, [ fail, eof ]"
        self.file_object.close()

    def exercise_partial_read(self):
        self.create_file_object(mode="rb")
        words = ext.partial_read(streambuf(self.file_object))
        assert words == b"Coding, should, "
        trailing = self.file_object.read()
        assert trailing == b" be fun"
        self.file_object.close()

    def exercise_read_failure(self):
        self.create_file_object(mode="rb")
        self.file_object.close()
        with pytest.raises(ValueError):
            ext.read_word(streambuf(self.file_object))
        self.file_object.close()

    def exercise_write(self):
        self.create_file_object(mode="wb")
        report = ext.write_word(ostream(self.file_object))
        assert report == b""
        assert self.file_content() == b"2 times 1.6 equals 3.2"
        self.file_object.close()

    def exercise_seek_and_read(self):
        self.create_file_object(mode="rb")
        instrumented_file = mock.Mock(spec=self.file_object, wraps=self.file_object)
        words = ext.read_and_seek(streambuf(instrumented_file))
        assert words == b"should, should, uld, ding, fun, [ eof ]"
        n = streambuf.default_buffer_size
        soughts = instrumented_file.seek.call_args_list
        # stringent tests carefully crafted to make sure the seek-in-buffer
        # optimisation works as expected
        # C.f. the comment in the C++ function actual_write_test
        assert soughts[-1] == mock.call(-4, 2)
        soughts = soughts[:-1]
        if n >= 14:
            assert soughts == []
        else:
            assert soughts[0] == mock.call(6, 0)
            soughts = soughts[1:]
            if 8 <= n <= 13:
                assert len(soughts) == 1 and self.only_seek_cur(soughts)
            elif n == 7:
                assert len(soughts) == 2 and self.only_seek_cur(soughts)
            else:
                assert soughts[0] == mock.call(6, 0)
                soughts = soughts[1:]
                assert self.only_seek_cur(soughts)
                if n == 4:
                    assert len(soughts) == 1
                else:
                    assert len(soughts) == 2
        self.file_object.close()

    def exercise_write_and_seek(self):
        self.create_file_object(mode="wb")
        instrumented_file = mock.Mock(spec=self.file_object, wraps=self.file_object)
        report = ext.write_and_seek(ostream(instrumented_file))
        assert report == b""
        expected = b"1000 times 1000 equals 1000000"
        assert self.file_content() == expected
        assert self.file_object.tell() == 9
        if streambuf.default_buffer_size >= 30:
            assert instrumented_file.write.call_count == 1
        self.file_object.close()

    @staticmethod
    def only_seek_cur(seek_calls):
        return all(call == mock.call(mock.ANY, 1) for call in seek_calls)


class bytesio_test_case(io_test_case):
    def exercise_write_failure(self):
        pass

    def create_file_object(self, mode):
        if mode == "rb":
            self.file_object = io.BytesIO(self.phrase)
        elif mode == "wb":
            self.file_object = io.BytesIO()
        else:
            raise NotImplementedError("Internal error in the test code")

    def file_content(self):
        return self.file_object.getvalue()


class mere_file_test_case(io_test_case):
    def exercise_write_failure(self):
        self.create_file_object(mode="rb")
        with pytest.raises(IOError):
            ext.write_word(streambuf(self.file_object))
        self.file_object.close()

    def create_file_object(self, mode):
        f = open("tmp_tst_python_streambuf", "wb")
        if mode.find("rb") > -1:
            f.write(self.phrase)
        f.close()
        self.file_object = open(f.name, mode)

    def file_content(self):
        self.file_object.flush()
        with open(self.file_object.name, "rb") as fh:
            result = fh.read()
        return result


def test_with_bytesio():
    bytesio_test_case().run()


@pytest.mark.xfail(
    "os.name == 'nt'", reason="crashes python process on Windows", run=False
)
def test_with_file(tmpdir):
    with tmpdir.as_cwd():
        mere_file_test_case().run()



from __future__ import annotations

from dials_algorithms_profile_model_modeller_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "CircleSampler",
    "EmpiricalProfileModeller",
    "EwaldSphereSampler",
    "GridSampler",
    "MultiExpProfileModeller",
    "ProfileModellerIface",
    "SamplerIface",
    "SingleSampler",
)



from __future__ import annotations

from libtbx.phil import parse

from dials.array_family import flex

phil_scope = parse(
    """
overlaps_filter {
  foreground_foreground {
    enable = False
      .type = bool
      .help = "Remove all spots in which neighbors' foreground"
              "impinges on the spot's foreground"
  }
  foreground_background {
    enable = False
      .type = bool
      .help = "Remove all spots in which neighbors' foreground"
              "impinges on the spot's background"
  }
}
""",
    process_includes=True,
)


class OverlapsFilter:
    from dials.algorithms.shoebox import MaskCode

    code_fgd = MaskCode.Foreground | MaskCode.Valid
    code_bgd = MaskCode.Background | MaskCode.Valid

    def is_fgd(self, code):
        return (code & self.code_fgd) == self.code_fgd

    def is_bgd(self, code):
        return (code & self.code_bgd) == self.code_bgd

    def __init__(self, refl, expt):
        self.refl = refl
        self.expt = expt
        self.masks = {}
        det = self.expt.detector
        assert len(det) == 1  # for now
        self.size_fast, self.size_slow = det[0].get_image_size()
        self.array_size = self.size_fast * self.size_slow

    def create_simple_mask(self):
        self.masks["simple_mask"] = flex.size_t(self.array_size)
        for obs in self.refl.rows():
            shoebox = obs["shoebox"]
            fast_coords = range(shoebox.xsize())
            slow_coords = range(shoebox.ysize())
            for f, s in zip(fast_coords, slow_coords):
                f_abs = f + shoebox.bbox[0]  # relative to detector
                s_abs = s + shoebox.bbox[2]  # relative to detector
                posn = f_abs + s_abs * self.size_fast  # position in mask array
                posn_in_shoebox = f + shoebox.xsize() * s  # position in shoebox
                try:
                    self.masks["simple_mask"][posn] |= shoebox.mask[posn_in_shoebox]
                except IndexError:  # bbox may extend past detector limits
                    continue

    def create_referenced_mask(self, test_code, mask_name):
        self.masks[mask_name] = [flex.size_t() for _ in range(self.array_size)]
        for idx in range(len(self.refl)):
            obs = self.refl[idx]
            shoebox = obs["shoebox"]
            fast_coords = range(shoebox.xsize())
            slow_coords = range(shoebox.ysize())
            for f, s in zip(fast_coords, slow_coords):
                f_abs = f + shoebox.bbox[0]  # relative to detector
                s_abs = s + shoebox.bbox[2]  # relative to detector
                posn = f_abs + s_abs * self.size_fast  # position in mask array
                posn_in_shoebox = f + shoebox.xsize() * s  # position in shoebox
                if (shoebox.mask[posn_in_shoebox] & test_code) == test_code:
                    try:
                        self.masks[mask_name][posn].append(idx)
                    except IndexError:  # bbox may extend past detector limits
                        continue

    def filter_using_simple_mask(self, mask_lambda, shoebox_lambda=lambda x: True):
        """At each pixel, examine the simple mask to determine if contributing
        observations should be excluded. When this condition mask_lambda is met,
        for each contributing observation, use optional condition shoebox_lambda
        to determine if the observation should be excluded (e.g. to exclude only
        those reflections contributing foreground). Return the mask reflecting
        this filter.
        """
        keep_refl_bool = flex.bool(len(self.refl), True)
        for idx in range(len(self.refl)):
            obs = self.refl[idx]
            shoebox = obs["shoebox"]
            fast_coords = range(shoebox.xsize())
            slow_coords = range(shoebox.ysize())
            for f, s in zip(fast_coords, slow_coords):
                f_abs = f + shoebox.bbox[0]  # relative to detector
                s_abs = s + shoebox.bbox[2]  # relative to detector
                posn = f_abs + s_abs * self.size_fast  # position in mask array
                posn_in_shoebox = f + shoebox.xsize() * s  # position in shoebox
                try:
                    if mask_lambda(
                        self.masks["simple_mask"][posn]
                    ):  # condition met in simple mask
                        if shoebox_lambda(
                            shoebox.mask[posn_in_shoebox]
                        ):  # condition met in shoebox
                            keep_refl_bool[idx] = False
                except IndexError:  # bbox may extend past detector limits
                    continue
        return keep_refl_bool

    def filter_all_using_referenced_mask(self, mask_name):
        """Return the mask reflecting the exclusion of any reflections for which the
        mask condition is true (e.g. untrusted pixels).
        """
        keep_refl_bool = flex.bool(len(self.refl), True)
        for i in self.masks[mask_name]:
            if len(i) > 0:
                for ref in i:
                    keep_refl_bool[ref] = False
        return keep_refl_bool

    def filter_overlaps_using_referenced_mask(self, mask_name):
        """At each pixel, define an overlap to be more than one reference (to an
        observation) indicated in the mask. Return the mask reflecting the exclusion
        of any overlaps (e.g. foreground with foreground).
        """
        keep_refl_bool = flex.bool(len(self.refl), True)
        for i in self.masks[mask_name]:
            if len(i) > 1:
                for ref in i:
                    keep_refl_bool[ref] = False
        return keep_refl_bool

    def remove_foreground_foreground_overlaps(self):
        self.create_referenced_mask(self.code_fgd, "foreground")
        self.refl = self.refl.select(
            self.filter_overlaps_using_referenced_mask("foreground")
        )

    def remove_foreground_background_overlaps(self):
        self.create_simple_mask()

        def is_overlap(code):
            return self.is_fgd(code) and self.is_bgd(code)

        self.refl = self.refl.select(
            self.filter_using_simple_mask(mask_lambda=is_overlap)
        )


class OverlapsFilterMultiExpt:
    def __init__(self, refl, expt):
        self.filters = [
            OverlapsFilter(r, e) for (r, e) in zip(refl.split_by_experiment_id(), expt)
        ]

    def remove_foreground_foreground_overlaps(self):
        for f in self.filters:
            f.remove_foreground_foreground_overlaps()

    def remove_foreground_background_overlaps(self):
        for f in self.filters:
            f.remove_foreground_background_overlaps()

    @property
    def refl(self):
        rlist = [f.refl for f in self.filters]
        r0 = flex.reflection_table()
        for r in rlist:
            r0.extend(r)
        return r0

    @property
    def expt(self):
        elist = [f.expt for f in self.filters]
        from dxtbx.model.experiment_list import ExperimentList

        e0 = ExperimentList()
        for e in elist:
            e0.append(e)
        return e0



#
#  Copyright (C) (2016) Lawrence Berkeley National Laboratory
#
#  Author: Nicholas Sauter.
#
#  This code is distributed under the BSD license, a copy of which is
#  included in the root directory of this package.
#

from __future__ import annotations

import logging

import libtbx
from scitbx.array_family import flex

from dials.algorithms.refinement import DialsRefineConfigError
from dials.algorithms.refinement.engine import AdaptLstbx as AdaptLstbxBase
from dials.algorithms.refinement.engine import DisableMPmixin
from dials.algorithms.refinement.engine import (
    GaussNewtonIterations as GaussNewtonIterationsBase,
)
from dials.algorithms.refinement.engine import LevenbergMarquardtIterations

try:
    from scitbx.examples.bevington import non_linear_ls_eigen_wrapper
except ImportError:
    raise DialsRefineConfigError(
        """Eigen package is not available.  Please untar the Eigen source package
     (http://eigen.tuxfamily.org) and place a link to it (eigen--> Eigen source dir) in
     the modules directory of your developer install; then recompile.
"""
    )

logger = logging.getLogger(__name__)


class AdaptLstbxSparse(DisableMPmixin, AdaptLstbxBase, non_linear_ls_eigen_wrapper):
    """Adapt the base class for Eigen"""

    def __init__(
        self,
        target,
        prediction_parameterisation,
        constraints_manager=None,
        log=None,
        tracking=None,
        max_iterations=None,
    ):

        AdaptLstbxBase.__init__(
            self,
            target,
            prediction_parameterisation,
            constraints_manager=constraints_manager,
            log=log,
            tracking=tracking,
            max_iterations=max_iterations,
        )

        non_linear_ls_eigen_wrapper.__init__(self, n_parameters=len(self.x))


class GaussNewtonIterations(AdaptLstbxSparse, GaussNewtonIterationsBase):
    """Refinery implementation, using lstbx Gauss Newton iterations"""

    def __init__(
        self,
        target,
        prediction_parameterisation,
        constraints_manager=None,
        log=None,
        tracking=None,
        max_iterations=20,
        **kwds,
    ):

        AdaptLstbxSparse.__init__(
            self,
            target,
            prediction_parameterisation,
            constraints_manager=constraints_manager,
            log=log,
            tracking=tracking,
            max_iterations=max_iterations,
        )

        # add an attribute to the journal
        self.history.add_column("reduced_chi_squared")  # flex.double()

        # adopt any overrides of the defaults above
        libtbx.adopt_optional_init_args(self, kwds)


class SparseLevenbergMarquardtIterations(
    GaussNewtonIterations, LevenbergMarquardtIterations
):
    """Levenberg Marquardt with Sparse matrix algebra"""

    def set_cholesky_factor(self):
        """Override that disables this method of the base AdaptLstbx. For
        sparse, large matrices this is numberically unstable; not to mention it
        is not implemented for the Eigen wrapper"""
        pass

    def setup_mu(self):
        """Override that works with the Eigen wrapper"""
        a_diag = self.get_normal_matrix_diagonal()
        self.mu = self.tau * flex.max(a_diag)

    def add_constant_to_diagonal(self, mu):
        """Delegate to the method of non_linear_ls_eigen_wrapper"""
        non_linear_ls_eigen_wrapper.add_constant_to_diagonal(self, self.mu)

    def report_progress(self, objective):
        """Override for the Eigen wrapper to provide live feedback of progress
        of the refinement"""

        logger.debug(
            "Iteration: %5d Objective: %18.4f Mu: %12.7f",
            self.n_iterations,
            objective,
            self.mu,
        )

    def run(self):
        self._run_core()

        # In contrast to dense matrix (traditional) LevMar, sparse matrix assumes
        # that the matrix is extremely large and not easily inverted. Therefore,
        # no attempt here to calculate esd's based on the variance covariance
        # matrix.

        logger.info(self.get_eigen_summary())
        return



from __future__ import annotations

import logging
import pickle

import iotbx.phil

import dials.util
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

logger = logging.getLogger("dials.command_line.find_hot_pixels")

phil_scope = iotbx.phil.parse(
    """\
output {
  mask = hot_pixels.pickle
    .type = path
}
""",
    process_includes=True,
)

help_message = """

  This program looks through the output of dials.find_spots to determine if any
  of the pixels are possible "hot" pixels. If a pixel is "hot" this will mean
  that it will have a consistently high value throughout the dataset. This
  program simply selects all pixels which are labelled as strong on all images
  in the dataset as "hot".

  Note that if you have still data or a small dataset, this is likely to produce
  lots of false positives; however, if you have a large rotation dataset, it is
  likely to be reasonably accurate.

  The program returns a file names hot_pixels.pickle which contains a boolean mask
  with True pixels being OK and False pixels being "hot" pixels.

  Examples::
    dials.find_hot_pixels models.expt strong.refl
"""


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util import Sorry, log

    usage = "dials.find_hot_pixels [options] models.expt strong.refl"

    # Create the option parser
    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    # Get the parameters
    params, options = parser.parse_args(args, show_diff_phil=False)

    # Configure the log
    log.config(verbosity=options.verbose, logfile="dials.find_hot_pixels.log")

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0 and len(reflections) == 0:
        parser.print_help()
        exit(0)

    if len(experiments) > 1:
        raise Sorry("Only one experiment can be processed at a time")
    else:
        imagesets = experiments.imagesets()
    if len(reflections) == 0:
        raise Sorry("No reflection lists found in input")
    if len(reflections) > 1:
        raise Sorry("Multiple reflections lists provided in input")

    assert len(reflections) == 1
    reflections = reflections[0]

    mask = hot_pixel_mask(imagesets[0], reflections)
    with open(params.output.mask, "wb") as fh:
        pickle.dump(mask, fh, pickle.HIGHEST_PROTOCOL)

    print(f"Wrote hot pixel mask to {params.output.mask}")


def hot_pixel_mask(imageset, reflections):
    depth = imageset.get_array_range()[1] - imageset.get_array_range()[0]
    xylist = filter_reflections(reflections, depth)

    from dials.array_family import flex

    mask = flex.bool(
        flex.grid(reversed(imageset.get_detector()[0].get_image_size())), True
    )

    for x, y in xylist:
        mask[y, x] = False

    print(f"Found {len(xylist)} hot pixels")

    return (mask,)


def filter_reflections(reflections, depth):
    xylist = []

    for i in range(len(reflections)):
        refl = reflections[i]
        s = refl["shoebox"]
        if s.zsize() == depth:
            xylist.append((s.xoffset(), s.yoffset()))

    return xylist


if __name__ == "__main__":
    run()



from __future__ import annotations

import logging.config
import os
import sys
import time
from typing import List

try:
    from colorlog import ColoredFormatter
except ImportError:
    ColoredFormatter = None

# https://stackoverflow.com/questions/25194864/python-logging-time-since-start-of-program/25196134#25196134
class DialsLogfileFormatter:
    """A formatter for log files that prepends messages with the elapsed time
    or messages at warning level or above with 'WARNING:'"""

    def __init__(self, timed):
        self.timed = timed
        self.start_time = time.time()
        self.prefix = ""

    def format(self, record):
        if self.timed:
            elapsed_seconds = record.created - self.start_time
            prefix = f"{elapsed_seconds:6.1f}: "
        else:
            prefix = ""
        indent = len(prefix)
        msg = record.getMessage()

        if record.levelno >= logging.WARNING:
            prefix = "{prefix:>{indent}s}".format(indent=indent, prefix="WARN: ")

        msg = msg.replace("\n", "\n" + " " * indent)
        if prefix == self.prefix:
            return " " * indent + msg
        else:
            self.prefix = prefix
            return prefix + msg


def config(verbosity=0, logfile=None):
    """
    Configure the logging.

    :param verbosity: Verbosity level of log output. Possible values:
                        * 0: Info log output to stdout/logfile
                        * 1: Info & debug log output to stdout/logfile
    :type verbosity: int
    :param logfile: Filename for log output.  If False, no log file is written.
    :type logfile: str
    """

    console = logging.StreamHandler(sys.stdout)
    if (
        "NO_COLOR" not in os.environ
        and sys.stdout.isatty()
        and ColoredFormatter is not None
    ):
        color_formatter = ColoredFormatter(
            "%(log_color)s%(message)s",
            log_colors={
                "DEBUG": "blue",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red,bg_white",
            },
        )
        console.setFormatter(color_formatter)

    dials_logger = logging.getLogger("dials")
    dials_logger.addHandler(console)
    dxtbx_logger = logging.getLogger("dxtbx")
    dxtbx_logger.addHandler(console)

    logging.captureWarnings(True)
    warning_logger = logging.getLogger("py.warnings")
    warning_logger.addHandler(console)

    if verbosity > 1:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.INFO

    if logfile:
        fh = logging.FileHandler(filename=logfile, mode="w", encoding="utf-8")
        fh.setLevel(loglevel)
        fh.setFormatter(DialsLogfileFormatter(timed=verbosity))
        dials_logger.addHandler(fh)
        dxtbx_logger.addHandler(fh)
        warning_logger.addHandler(fh)

    dials_logger.setLevel(loglevel)
    dxtbx_logger.setLevel(loglevel)
    console.setLevel(loglevel)

    print_banner(use_logging=True)


class CacheHandler(logging.Handler):
    """A simple class to store log messages."""

    def __init__(self):
        """
        Initialise the handler
        """
        super().__init__()
        self.records = []

    def emit(self, record):
        """
        Emit the message to a list

        :param record: The log record
        """
        self.records.append(record)


def config_simple_cached():
    """
    Configure the logging to use a cache.
    """

    # Configure the logging
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "handlers": {
                "cache": {"level": "DEBUG", "class": "dials.util.log.CacheHandler"}
            },
            "loggers": {
                "dials": {"handlers": ["cache"], "level": "DEBUG", "propagate": True}
            },
        }
    )


def rehandle_cached_records(records: List[logging.LogRecord]) -> None:
    """
    Submit cached log records to the relevant loggers for handling.

    Because the relevant logger's threshold log level may be higher than when the
    original log message was created and cached, the record will only be re-handled
    if its level meets the new threshold.

    Args:
        records:  Cached log records.
    """
    for record in records:
        record_logger = logging.getLogger(record.name)
        if record_logger.isEnabledFor(record.levelno):
            record_logger.handle(record)


_banner = (
    "DIALS (2018) Acta Cryst. D74, 85-97. https://doi.org/10.1107/S2059798317017235"
)
_banner_printed = False


def print_banner(force=False, use_logging=False):
    global _banner_printed
    if _banner_printed and not force:
        return
    if os.getenv("DIALS_NOBANNER"):
        return
    _banner_printed = True

    if use_logging:
        logging.getLogger("dials").info(_banner)
    else:
        print(_banner)


class LoggingContext:
    # https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging
    def __init__(self, logger, level=None):
        self.logger = logging.getLogger(logger) if isinstance(logger, str) else logger
        self.level = level

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        # implicit return of None => don't swallow exceptions



from __future__ import annotations


class H5PYEncoder:
    """Encoder base class."""

    def encode(self, obj, handle):
        """Encode the object in the HDF5 file."""
        raise RuntimeError("Overload!")


class H5PYDecoder:
    """Decoder base class."""

    def decode(self, handle):
        """Decode the object from the HDF5 file."""
        raise RuntimeError("Overload!")


class ReflectionListEncoder(H5PYEncoder):
    """Encoder for the reflection data."""

    def encode(self, reflections, handle):
        """Encode the reflection data."""

        # Create the reflection data group
        group = handle.create_group("entry/data_processing")
        group.attrs["num_reflections"] = len(reflections)

        # Encode each column entry
        for key, data in reflections.cols():
            self.encode_column(group, key, data)

    def encode_column(self, group, key, data):
        """Encode a column of data."""
        from dials.array_family import flex

        if isinstance(data, flex.shoebox):
            self.encode_shoebox(group, key, data)
        else:
            group[key] = list(data)
            group[key].attrs["flex_type"] = type(data).__name__

    def encode_shoebox(self, group, key, sb_data):
        """Encode a column of shoeboxes."""
        shoebox = group.create_group("shoebox")
        shoebox.attrs["flex_type"] = type(sb_data).__name__
        data = shoebox.create_group("data")
        mask = shoebox.create_group("mask")
        background = shoebox.create_group("background")
        for i, sb in enumerate(sb_data):
            data["%d" % i] = sb.data.as_numpy_array()
            mask["%d" % i] = sb.mask.as_numpy_array()
            background["%d" % i] = sb.background.as_numpy_array()


class ReflectionListDecoder(H5PYDecoder):
    """Decoder for the reflection data."""

    def decode(self, handle):
        """Decode the reflection data."""
        from dials.array_family import flex

        # Get the group containing the reflection data
        g = handle["entry/data_processing"]

        # Create the list of reflections
        rl = flex.reflection_table(int(g.attrs["num_reflections"]))

        # Decode all the columns
        for key in g:
            item = g[key]
            name = item.attrs["flex_type"]
            if name == "shoebox":
                flex_type = getattr(flex, name)
                data = item["data"]
                mask = item["mask"]
                background = item["background"]
                col = flex_type(len(rl))
                for i in range(len(rl)):
                    col[i].data = flex.double(data["%d" % i].value)
                    col[i].mask = flex.int(mask["%d" % i].value)
                    col[i].background = flex.double(background["%d" % i].value)

            else:
                flex_type = getattr(flex, name)
                col = self.decode_column(flex_type, item)
            rl[str(key)] = col

        # Return the list of reflections
        return rl

    def decode_column(self, flex_type, data):
        """Decode a column for various flex types."""
        from dials.array_family import flex

        if flex_type == flex.int:
            return flex_type(int(d) for d in list(data))
        elif flex_type == flex.double:
            return flex_type(float(d) for d in list(data))
        elif flex_type == flex.vec3_double:
            return flex_type(list(map(float, d)) for d in list(data))
        elif flex_type == flex.vec2_double:
            return flex_type(list(map(float, d)) for d in list(data))
        elif flex_type == flex.int6:
            return flex_type(list(map(int, d)) for d in list(data))
        elif flex_type == flex.miller_index:
            return flex_type(list(map(int, d)) for d in list(data))
        elif flex_type == flex.bool:
            return flex_type(bool(d) for d in list(data))
        elif flex_type == flex.size_t:
            return flex_type(int(d) for d in list(data))
        else:
            return flex_type(list(data))


class NexusFile:
    """Interface to Nexus file."""

    def __init__(self, filename, mode="a"):
        """Open the file with the given mode."""
        import h5py

        self._handle = h5py.File(filename, mode)

    def close(self):
        """Close the file."""
        self._handle.close()
        del self._handle

    def set_data(self, data, encoder):
        """Set the model data using the supplied encoder."""
        encoder.encode(data, self._handle)

    def get_data(self, decoder):
        """Get the model data using the supplied decoder."""
        return decoder.decode(self._handle)

    def set_reflections(self, reflections):
        """Set the reflection data."""
        self.set_data(reflections, ReflectionListEncoder())

    def get_reflections(self):
        """Get the reflection data."""
        return self.get_data(ReflectionListDecoder())


if __name__ == "__main__":
    from dials.array_family import flex

    reflections = flex.reflection_table(
        [
            ("hkl", flex.miller_index(10)),
            ("s1", flex.vec3_double(10)),
            ("bbox", flex.int6(10)),
            ("id", flex.int(10)),
            ("shoebox", flex.shoebox(10)),
        ]
    )

    for i in range(10):
        reflections["shoebox"][i].data = flex.double(flex.grid(10, 10, 10))
        reflections["shoebox"][i].mask = flex.int(flex.grid(10, 10, 10))
        reflections["shoebox"][i].background = flex.double(flex.grid(10, 10, 10))

    for i in range(10):
        print(reflections["shoebox"][i].data.all())

    writer = NexusFile("temp.h5", "w")
    writer.set_reflections(reflections)
    writer.close()

    reader = NexusFile("temp.h5", "r")
    reflections = reader.get_reflections()
    reader.close()

    for r in reflections:
        print(r)
